import pathlib
from .utilities import get_all_files, dicom_to_numpy
import pydicom
from pathlib import Path
import numpy as np
from skimage.filters import gaussian, threshold_otsu
from skimage.measure import label
from scipy.interpolate import NearestNDInterpolator
from matplotlib import pyplot as plt
import logging
import os
import json
import pandas as pd
from scipy.spatial.distance import cdist
from scipy.spatial import transform
from datetime import datetime
from warnings import warn

ch = logging.StreamHandler()
formatter = logging.Formatter('[%(filename)s: line %(lineno)d %(levelname)8s] %(message)s')
ch.setFormatter(formatter)
logger = logging.getLogger(__name__)
logger.addHandler(ch)
logger.setLevel(logging.INFO)
logger.propagate = False


class MarkerVolume:
    """
    This class excepts either a path to a folder of dicom files or a numpy array of marker positions.
    In the former case, it will attempt to automatically extract the marker positions. In the latter, it will just
    use the input marker positions. Once the instance is created, it contains a number of handy plots and functions.

    :param input_data_path: Either a path to a folder of dicoms, or a numpy array with three columns corresponding to the
        [x,y,z] marker positions
    :type input_data_path: path, string, np.ndarray
    :param ImExtension: change to whatever extension your dicom files, most commonly either dcm or IMA
    :type ImExtension: str, optional
    :param r_min: any markers at radii less than this will be discounted.
    :type r_min: float ,optional
    :param r_max: any markers at radii greater than this will be discounted.
    :type r_max: float, optional
    :param iterative_segmentation: If set to true, a slower iterative method will be used to find the threshold value
        used to segment the markers. Otherwise, otsu's method is used.
    :type iterative_segmentation: bool, optional
    :param threshold: Manually set the threshold value. If this is left as None otsu's method is used.
        This will replace the iterative segmentation method regardless of flag.
    :type threshold: float, optional
    :param n_markers_expected: if you know how many markers you expect, you can enter the value here. The code will
        then warn you if it finds a different number.
    :type n_markers_expected: int, optional
    :param gaussian_image_filter_sd: size of filter to blur input data with. Higher values are better at removing
        noise prior to thresholding, but may also blur out small/low signal markers!
    :type gaussian_image_filter_sd: float, optional
    :param correct_fat_water_shift: if True, will attempt to automatically correct marker positions for calculated
        fat water shift
    :type correct_fat_water_shift: bool, optional
    :param marker_size_lower_tol: lower acceptance window for a marker compared to the median size of markers.
        e.g. markermarker_size_tol=0.5 will result in markers being accepted within when
        (1-marker_size_lower_tol)*median_marker_size <= marker_size <= (1+markermarker_size_tol)*marker_size_upper_tol
    :type marker_size_lower_tol: float, optional
    :param marker_size_upper_tol: upper acceptance window for a marker compared to the median size of markers.
    :type marker_size_upper_tol: float, optional
    :param verbose: if True, prints a lot of info about the marker extraction process to the screen. mostly useful
        during debugging
    :type verbose: bool, optional
    :param fat_shift_direction: We aren't sure yet if we can confidently predict the direction that the fat water shift
        will be in. if you set correct_fat_water_shift=True, you can change the direction it shifts in here. it should
        take a value of -1, or 1. pull requests making this easier to use very welcome!
    :type fat_shift_direction: int, optional
    """

    def __init__(self, input_data, ImExtension='dcm', r_min=None, r_max=None, iterative_segmentation=False,
                 threshold=None, n_markers_expected=None, fat_shift_direction=None, verbose=False,
                 gaussian_image_filter_sd=1, correct_fat_water_shift=False, marker_size_lower_tol=0.9,
                 marker_size_upper_tol=1):

        self.verbose = verbose
        self._file_extension = ImExtension
        self._correct_fat_water_shift = correct_fat_water_shift
        self.dicom_data = None  # overwritten below if possible.
        self._n_markers_expected = n_markers_expected
        self._r_min = r_min
        self._r_max = r_max
        self._cutoffpoint = threshold
        self._iterative_segmentation = iterative_segmentation
        self._chemical_shift_vector = None
        self._gaussian_image_filter_sd = gaussian_image_filter_sd
        self._marker_size_lower_tol = marker_size_lower_tol
        self._marker_size_upper_tol = marker_size_upper_tol

        if isinstance(input_data, (pathlib.Path, str)):
            self.input_data_path = Path(input_data)
            if self.input_data_path.is_dir():
                # dicom input
                self.InputVolume, self.dicom_affine, (self.X, self.Y, self.Z) = \
                    dicom_to_numpy(self.input_data_path, file_extension=self._file_extension, return_XYZ=True)
                self._calculate_MR_acquisition_data()

                # Segmenting markers
                self._filter_volume_by_r()
                self.ThresholdVolume, self.BlurredVolume = self._threshold_volume(self.InputVolume)
                centroids = self._find_contour_centroids()
                self.MarkerCentroids = pd.DataFrame(centroids, columns=['x', 'y', 'z'])

                # Correct for oil water shift
                if self._correct_fat_water_shift:
                    self._calculate_chemical_shift_vector(fat_shift_direction=fat_shift_direction)
                    self.MarkerCentroids = self.MarkerCentroids + self._chemical_shift_vector
            elif (os.path.isfile(self.input_data_path) and os.path.splitext(self.input_data_path)[1] == '.json'):
                # slicer input
                with open(self.input_data_path) as f:
                    data = json.load(f)
                arrays = [el for el in data['markups'][0]['controlPoints']]
                points = [el['position'] for el in arrays]
                self.MarkerCentroids = pd.DataFrame(points, columns=['x', 'y', 'z'])

                # look for dicom_data json file
                try:
                    with open(os.path.join(self.input_data_path.parent, 'dicom_data.json')) as f:
                        self.dicom_data = json.load(f)
                except:
                    logger.warning(f'MR data file dicom_data.json not found at {self.input_data_path.parent}. Continuing')
            else:
                raise FileNotFoundError(f'could not find any data at {self.input_data_path}')
        elif isinstance(input_data, np.ndarray):
            # don't need to do anything then.
            self.MarkerCentroids = pd.DataFrame(input_data, columns=['x', 'y', 'z'])
            logger.warning("numpy input is deprecated, please use pandas input with ['x', 'y', 'z'] cols instead")
        elif isinstance(input_data, pd.core.frame.DataFrame):
            _expected_columns = ['x', 'y', 'z']
            if not all([col in input_data for col in _expected_columns]):
                raise AttributeError("input pandas data must have columns called ['x', 'y', 'z']")
            self.MarkerCentroids = input_data.copy()
        else:
            raise TypeError(f'Uknown data input: {type(input_data)}')

        # insert the radial value of each marker:
        self.MarkerCentroids['r'] = self.MarkerCentroids.apply(
            lambda row: np.sqrt(row[0] ** 2 + row[1] ** 2 + row[2] ** 2), axis=1)
        if self.MarkerCentroids.r.mean() < 10:
            logger.warning('it appears that your input data is in m, not mm - please use mm!')

        self._filter_markers_by_r()
        if self._n_markers_expected is not None:
            if not (self.MarkerCentroids.shape[0] == n_markers_expected):
                logger.warning(f'For data {self.input_data_path}\n'
                               f'You entered that you expected to find'
                               f' {n_markers_expected}, but actually found {self.MarkerCentroids.shape[0]}.')

    def _filter_volume_by_r(self):
        """
        Accelerates and improves segmenting by setting all voxels outside of specified radius to the background
        """
        if self._r_max or self._r_min:
            _max = 10000
            _min = -1
            if self._r_max:
                _max = self._r_max
            if self._r_min:
                _min = self._r_min
            background = np.median(self.InputVolume)

            it = np.nditer(self.InputVolume, flags=['multi_index'])
            for voxel in it:
                distance = np.sqrt(
                    self.X[it.multi_index] ** 2 + self.Y[it.multi_index] ** 2 + self.Z[it.multi_index] ** 2)
                if distance < _min or distance > _max:
                    self.InputVolume[it.multi_index] = background

    def _filter_markers_by_r(self):
        """
        Remove any markers that are less than r_min or more than r_max
        """
        if self._r_max:
            keep_ind = abs(self.MarkerCentroids.r) < self._r_max
            self.MarkerCentroids = self.MarkerCentroids[keep_ind].reset_index(drop=True)
        if self._r_min:
            keep_ind = abs(self.MarkerCentroids.r) > self._r_min
            self.MarkerCentroids = self.MarkerCentroids[keep_ind].reset_index(drop=True)

    def _calculate_MR_acquisition_data(self):
        """
        If we have MR images, extract the information from thall_dicom_files = get_all_files(self.input_data_path)e
        dicom header that is needed to convert marker positions to magnetic fields downstream.
        - This method is dependant on having the full coordinate list which are calculated upstream; this is why it
            must reside inside this class.
        - We include many fields that aren't in the dicom header by default
        - we package this as a simple dict that can be easily saved as json
        """
        all_dicom_files = get_all_files(self.input_data_path, file_extension=self._file_extension)
        # use first one, its as good as any other:
        example_dicom_file = pydicom.read_file(self.input_data_path / all_dicom_files[0])

        if example_dicom_file.Modality == 'MR':
            if not example_dicom_file.Manufacturer == 'SIEMENS':
                logger.warning(
                    'this code has not been tested with non-siemens scanners. If dicom standards work properly'
                    'this code will too...')
            x = np.unique(self.X)
            y = np.unique(self.Y)
            z = np.unique(self.Z)
            x_pixel_spacing = np.mean(np.diff(x))
            y_pixel_spacing = np.mean(np.diff(y))
            z_pixel_spacing = np.mean(np.diff(z))
            x_fov = x.max() - x.min() + x_pixel_spacing
            y_fov = y.max() - y.min() + y_pixel_spacing
            z_fov = z.max() - z.min() + z_pixel_spacing
            self._dicom_header = example_dicom_file

            self.dicom_data = {}  # fill up below
            self.dicom_data['FOV'] = [x_fov, y_fov, z_fov]
            self.dicom_data['bandwidth'] = float(example_dicom_file.PixelBandwidth)

            self.dicom_data['gama'] = float(example_dicom_file.MagneticFieldStrength * 42.57747851)
            self.dicom_data['pixel_spacing'] = [x_pixel_spacing, y_pixel_spacing, z_pixel_spacing]
            self.dicom_data['image_size'] = [x.size, y.size, z.size]
            self.dicom_data['magnetic_field_strength'] = example_dicom_file.MagneticFieldStrength
            self.dicom_data['imaging_frequency'] = example_dicom_file.ImagingFrequency
            try:
                acquisition_date = datetime.strptime(example_dicom_file.AcquisitionDate, '%Y%m%d')
                self.dicom_data['acquisition_date'] = acquisition_date.strftime("%d_%B_%Y")
            except ValueError as e:
                logger.error(e)
                logger.error('failed to extract datetime, probably didnt understand date format, continuing...')
            self.dicom_data['manufacturer'] = example_dicom_file.Manufacturer

            # the above works on siemens scanner, not sure if it will prove consistent on others
            fat_water_delta_f = example_dicom_file.MagneticFieldStrength * 42.57747851 * 3.5
            self.dicom_data['chem_shift_magnitude'] = (example_dicom_file.PixelBandwidth / fat_water_delta_f) / 2
            self.dicom_data['InPlanePhaseEncodingDirection'] = example_dicom_file.InPlanePhaseEncodingDirection
            # extract the frequency, phase, and slice directions (B0 effects occur in freq and slice)
            directions = ['x', 'y', 'z']
            phase_encode_direction = None
            freq_encode_direction = None
            if self.dicom_data['InPlanePhaseEncodingDirection'] == 'ROW':
                phase_cosines = example_dicom_file.ImageOrientationPatient[:3]
                row_cosines = example_dicom_file.ImageOrientationPatient[3:]
            elif self.dicom_data['InPlanePhaseEncodingDirection'] == 'COL':
                phase_cosines = example_dicom_file.ImageOrientationPatient[3:]
                row_cosines = example_dicom_file.ImageOrientationPatient[:3]
            for i, (phase_cosine, freq_cosine) in enumerate(zip(phase_cosines, row_cosines)):
                if abs(float(phase_cosine)) == 1:
                    phase_encode_direction = directions[i]
                if abs(float(freq_cosine)) == 1:
                    freq_encode_direction = directions[i]
            if phase_encode_direction is None:
                logger.warning('failed to extract phase encoding direction from dicom data, continuing')
            if freq_encode_direction is None:
                logger.warning('failed to extract frequency encoding direction from dicom data, continuing')
            self.dicom_data['phase_encode_direction'] = phase_encode_direction
            self.dicom_data['freq_encode_direction'] = freq_encode_direction
            # get slice direction
            slice_dir_ind = np.equal(example_dicom_file.ImageOrientationPatient[:3],
                                     example_dicom_file.ImageOrientationPatient[3:])
            self.dicom_data['slice_direction'] = np.array(directions)[slice_dir_ind][0]
            bandwidth = np.array(self.dicom_data['bandwidth'])
            image_size = np.array(self.dicom_data['image_size'])
            gama = np.array(self.dicom_data['gama'])
            FOV = np.array(self.dicom_data['FOV'])
            gradient_strength = bandwidth * image_size / (gama * 1e6 * FOV * 1e-3)  # unit(T / m)
            self.dicom_data['gradient_strength'] = list(gradient_strength)

        else:
            self.dicom_data = None

    def _calculate_chemical_shift_vector(self, fat_shift_direction=1):
        """
        Calculates a chemical shift vector (x,y,z) to be applied to each marker to account for the fat/water shift.

        :param fat_shift_direction: dictates whether the shift is positive or negative. we aren't sure how consisten
            this is between scanners so have left as used defined.
        """
        if self.dicom_data is None:
            logger.warning('No dicom data for chemical shift')
            return

        # TODO check these
        if fat_shift_direction == 'AP' or fat_shift_direction == 'RL' or fat_shift_direction == 'FH' or fat_shift_direction == -1:
            fat_shift_direction = -1
        elif fat_shift_direction == 'PA' or fat_shift_direction == 'LR' or fat_shift_direction == 'HF' or fat_shift_direction == 1:
            fat_shift_direction = 1
        else:
            fat_shift_direction = 0
            logger.warning('Chemical shift input was invalid')

        # Calculate magnitude (in pixels) of the chemical shift based on the shift frequency of a 1T MRI field (149Hz)
        directions = ['x', 'y', 'z']
        direction = np.array(([int(el == self.dicom_data['freq_encode_direction']) for el in directions]))
        # ^ this just gives a vector with a 1 in the correct direction e.g. [0,1,0]
        shift = direction * self.dicom_data['chem_shift_magnitude']
        self._chemical_shift_vector = shift * np.array(self.dicom_data['pixel_spacing'])[
            direction.astype(bool)] * fat_shift_direction

    def _find_iterative_cutoff(self, blurred_volume):
        """
        Replaces the otsu threshold with a slower, iterative method to find the best threshold for segmenting markers.
        A range of valid thresholds are found and the mean is returned as the best threshold.
        If no valid thresholds are found, the mean of candidate thresholds that will miss some markers is returned.
        """

        if self._n_markers_expected is None:
            logger.warning('Iterative segmentation method requires the expected number of markers to be input.')
            return None

        # finds a range of thresholds that give a number of segments near to the number of expected markers
        histogram_division = 100

        candidate_thresholds = []       # Thresholds that have fewer markers than expected, use when no valid thresholds
        candidate_n_points = []
        valid_thresholds = []           # Thresholds that should give the corrected number of markers

        # divide the range into segments and calculate how many volumes result from each segment
        background_value = np.median(np.round(blurred_volume))
        intensity_range = np.max(blurred_volume) - background_value
        histogram_segment = intensity_range / histogram_division

        # Testing each histogram segment
        for i in range(1, histogram_division - 1):

            n_segments = 0
            cutoff = background_value + i * histogram_segment
            threshold_volume = blurred_volume > cutoff
            labels = label(threshold_volume, background=0)
            unique_labels = np.unique(labels)[1:]  # first label is background so skip
            # Check number of segments falls within valid range:
            tol = 0.1
            if len(unique_labels) < self._n_markers_expected*(1-tol) or len(unique_labels) > self._n_markers_expected * (1+tol):
                continue  # Too few or too many segments
            else:
                # This threshold is possibly valid
                candidate_thresholds.append(cutoff)
                # Remove small volumes and check if still valid
                for label_level in unique_labels:
                    RegionInd = labels == label_level
                    if np.count_nonzero(RegionInd) > 2:
                        n_segments += 1
                candidate_n_points.append(n_segments)
                if n_segments >= self._n_markers_expected:
                    valid_thresholds.append(cutoff)
                    if self.verbose:
                        print('Threshold: ' + str(format(cutoff, '.2f')) + ', ' + str(n_segments) + ' segments.')
                else:
                    if self.verbose:
                        print('Threshold: ' + str(format(cutoff, '.2f')) + ' is a candidate.')

            if len(unique_labels) < 10:
                # threshold is too high and we can stop
                break

        if len(valid_thresholds) > 0:
            return np.mean(valid_thresholds)
        elif len(candidate_thresholds) > 0:
            ind = np.argmin(abs(np.subtract(self._n_markers_expected, candidate_n_points)))
            best_threshold = candidate_thresholds[ind]
            logger.warning(f'No valid thresholds were found. Using closest possible value of {best_threshold}')
            return best_threshold
        else:
            logger.warning('No valid thresholds were found. Try lowering gaussian blur level.')
            return None

    def _threshold_volume(self, VolumeToThreshold):
        """
        For a 3D numpy array, turn all elements < cutoff to zero, and all other elements to 1.
        If no cutoff is entered, otsu's method is used to auto-threshold.
        """

        BlurredVolume = gaussian(VolumeToThreshold, sigma=self._gaussian_image_filter_sd)

        if self._cutoffpoint is not None:
            # If cutoff point has been manually entered, go straight to thresholding
            self._iterative_segmentation = False
        if self._iterative_segmentation is True:
            self._cutoffpoint = self._find_iterative_cutoff(BlurredVolume)
        if self._cutoffpoint is None:
            self._cutoffpoint = threshold_otsu(BlurredVolume)

        ThresholdVolume = BlurredVolume > self._cutoffpoint

        return ThresholdVolume, BlurredVolume

    def _get_marker_max_min_volume(self):
        """
        figures out the range of voxels that a marker should have
        """
        self._n_voxels = []  # going to keep track of this so we can remove any very small regions if needed

        for label_level in self.unique_labels:  # first label is background so skip
            # extract x,y,z of each connected region
            RegionInd = self._labels == label_level
            self._n_voxels.append(np.count_nonzero(RegionInd))
        # Set up min and max marker volumes
        self._n_voxels_median = np.median(np.array(self._n_voxels))
        self._voxel_min = (1 - self._marker_size_lower_tol) * self._n_voxels_median
        self._voxel_max = (1 + self._marker_size_upper_tol) * self._n_voxels_median

        # Modify based on number of markers
        if self._n_markers_expected is not None:
            self._n_voxels_copy = np.array(self._n_voxels)
            voxel_difference = np.absolute(np.array(self._n_voxels) - self._n_voxels_median)
            # Loop that deletes the marker size with the largest difference from median until the expected # of markers
            while len(self._n_voxels_copy) > self._n_markers_expected:
                self._n_voxels_copy = np.delete(self._n_voxels_copy, np.argmax(voxel_difference))
                voxel_difference = np.delete(voxel_difference, np.argmax(voxel_difference))
            # Set min and max based on the remaining list
            if np.min(self._n_voxels_copy) > self._voxel_min:
                self._voxel_min = np.min(self._n_voxels_copy)
            if np.max(self._n_voxels_copy) < self._voxel_max:
                self._voxel_max = np.max(self._n_voxels_copy)

        # 3 voxels is the absolute floor
        if self._voxel_min < 3:
            self._voxel_min = 3

    def _remove_load(self):
        '''
        if it seems very likely that the only thing that has been segmented is the load
        remove it and try again. We may want to enable this in situations where any large object is detected...
        '''

        for label_level in self.unique_labels:
            RegionInd = self._labels == label_level
            self.InputVolume[RegionInd] = 0
        self._cutoffpoint = None
        self.ThresholdVolume, self.BlurredVolume = self._threshold_volume(self.InputVolume)
        self._labels = label(self.ThresholdVolume, background=0)
        self.unique_labels = np.unique(self._labels)[1:]  # first label is background so skip

    def _print_thresholding_information(self):
        print('Threshold value: ' + str(self._cutoffpoint))
        print('Median marker volume: ' + str(self._n_voxels_median))
        print('Expected marker volume: ' + str(self._voxel_min) + ' to ' + str(self._voxel_max))
        print(f'Regions to check: {self.unique_labels.shape}')

    def _print_thresholding_summary(self, n_found_markers, n_skipped_markers):
        print('----------')
        print(f'Total regions: {len(self.unique_labels)}')
        print(f'Total markers found:   {n_found_markers}')
        print(f'Total skipped:         {n_skipped_markers}')

    def _find_contour_centroids(self):
        """
        This code loops through all the found regions, extracts the cartesian coordiantes, and takes the
        intensity-weighted average as the centroid.
        It excludes volumes that are bigger/ smaller than the median using the params
        marker_size_upper_tol and marker_size_lower_tol
        """
        self._labels = label(self.ThresholdVolume, background=0)
        self.unique_labels = np.unique(self._labels)[1:]  # first label is background so skip
        if self.unique_labels.shape[0] < 3:
            logger.warning('automatic thresholding didnt work, trying to remove load and try again...')
            self._remove_load()
        self._get_marker_max_min_volume()

        if self.verbose:
            self._print_thresholding_information()

        x_centroids = []
        y_centroids = []
        z_centroids = []
        skipped = 0

        # extract x,y,z of each connected region
        for i, label_level in enumerate(self.unique_labels):
            RegionInd = np.equal(self._labels, label_level)
            voxels = np.count_nonzero(RegionInd)
            if voxels < self._voxel_min or voxels > self._voxel_max:
                skipped += 1
                if self.verbose:
                    print('Region ' + str(i + 1) + ': Skipped, v = ' + str(voxels))
                continue  # skip outliers

            region_sum = np.sum(self.InputVolume[RegionInd])
            weighted_x = np.sum(np.multiply(self.X[RegionInd], self.InputVolume[RegionInd]))
            weighted_y = np.sum(np.multiply(self.Y[RegionInd], self.InputVolume[RegionInd]))
            weighted_z = np.sum(np.multiply(self.Z[RegionInd], self.InputVolume[RegionInd]))
            x_centroids.append(weighted_x / region_sum)
            y_centroids.append(weighted_y / region_sum)
            z_centroids.append(weighted_z / region_sum)
            if self.verbose:
                print('Region ' + str(i + 1) + ': Marker, v = ' + str(voxels))

        if self.verbose:
            self._print_thresholding_summary(len(x_centroids), skipped)

        return np.array([x_centroids, y_centroids, z_centroids]).T

    # public methods

    def plot_3D_markers(self, title='3D marker positions'):  # pragma: no cover
        """
        Just a quick plot of the marker positions; very useful sanity check!
        """

        fig = plt.figure()
        axs = fig.add_subplot(111, projection='3d')
        axs.scatter(self.MarkerCentroids.x, self.MarkerCentroids.y, self.MarkerCentroids.z, s=10)
        axs.set_xlabel('X [mm]')
        axs.set_ylabel('Y [mm]')
        axs.set_zlabel('Z [mm]')
        axs.set_title(title)
        axs.set_box_aspect(
            (np.ptp(self.MarkerCentroids.x), np.ptp(self.MarkerCentroids.y), np.ptp(self.MarkerCentroids.z)))
        plt.show()

    def perturb_marker_positions(self, random_perturbation, systemic_perturbation=0):
        """
        add random noise to the marker positions. Useful to perform sensitivity analysis. operates in place.

        :param random_perturbation: upper and lower limits of random_perturbation to each marker coordinates. uniform
            distribution is used
        :type random_perturbation: float
        :param systemic_perturbation: constant offset added to all markers
        :type systemic_perturbation: float, optional
        """

        x_peturb = np.random.uniform(low=-random_perturbation, high=random_perturbation,
                                     size=self.MarkerCentroids.x.shape)
        y_peturb = np.random.uniform(low=-random_perturbation, high=random_perturbation,
                                     size=self.MarkerCentroids.x.shape)
        z_peturb = np.random.uniform(low=-random_perturbation, high=random_perturbation,
                                     size=self.MarkerCentroids.x.shape)
        self.MarkerCentroids.x = self.MarkerCentroids.x + x_peturb + systemic_perturbation
        self.MarkerCentroids.y = self.MarkerCentroids.y + y_peturb + systemic_perturbation
        self.MarkerCentroids.z = self.MarkerCentroids.z + z_peturb + systemic_perturbation

    def export_to_slicer(self, save_path=None, filename='slicer_centroids'):
        """
        export a json file that can be read in by slicer, allowing a good way to visualise marker segmentation
        performance. This file will be saved at the same spot as input data if dicom was input, otherwise it will be
        saved to save_path.
        """
        # sort out where to save:
        if save_path is None:
            try:
                _save_path = self.input_data_path
            except AttributeError:  # indicates numpy input
                logger.error('To save numpy data to slicer format, please enter a value for "save_path" when'
                             'calling MarkerVolume.export_to_slicer')
                return
        else:
            _save_path = Path(save_path)

        file_content = []  # will fill up one line at a time
        file_content.append('{"@schema": "https://raw.githubusercontent.com/slicer/slicer/master/Modules/Loadable/'
                            'Markups/Resources/Schema/markups-schema-v1.0.0.json#",\n')
        file_content.append('"markups": [{"type": "Fiducial", "coordinateSystem": "LPS", "controlPoints": [\n')
        for index, row in self.MarkerCentroids.iterrows():
            if index == self.MarkerCentroids.shape[0] - 1:
                # for last entry have to delete a comma
                file_content.append(
                    f'{{ "label": " ", "position": [{row.x: 1.2f}, {row.y: 1.2f}, {row.z: 1.2f}], "locked": true}}\n')
            else:
                file_content.append(
                    f'{{ "label": " ", "position": [{row.x: 1.2f}, {row.y: 1.2f}, {row.z: 1.2f}], "locked": true}},\n')

        file_content.append(']}]}')
        # write contents
        filename = filename + '.mrk.json'
        file_loc = _save_path / filename
        with open(file_loc, 'w') as f:
            f.writelines(file_content)

    def save_dicom_data(self, save_path=None, filename='dicom_data'):
        """
        save the dicom data as json.  This file will be saved at the same
        spot as input data if dicom was input, otherwise it will be saved
        to save_path.
        """

        if self.dicom_data is None:
            logger.warning('cannot save dicom data because there is none...')
            return
        if save_path is None:
            save_path = self.input_data_path
        save_path = Path(save_path)
        _file_name, _file_extension = os.path.splitext(filename)
        if _file_extension == '':
            _file_extension = '.json'
        filename = _file_name + _file_extension
        full_filename = save_path / filename

        with open(full_filename, 'w') as f:
            json.dump(self.dicom_data, f)


class MatchedMarkerVolumes:
    """
    :param GroundTruthData:
    :type GroundTruthData: MarkerVolume object, or path to dicom folder, or numpy array
    :param DistortedData:
    :type DistortedData: MarkerVolume object, or path to dicom folder, or numpy array
    :param reverse_gradient_data:
    :type reverse_gradient_data: None, or MarkerVolume object, or path to dicom folder, or numpy array
    :param WarpSearchData: if True, position of found markers is used to update expected positions for ground truth.
        Recomended is True if there is substantial distortion present.
    :type WarpSearchData: boolean
    :param AutomatchMarkers: if True, will automatically match the distorted data to the ground truth using a search
        algorith. If False, will simply use input data in the order it is entered. This is useful if you have sets
        of markers that were independently matched through some other process.
    :type AutomatchMarkers: boolean
    :param AllowDoubleMatching: when False, each ground truth marker is only allowed to match with one distorted
        marker. However, sometimes this can result in a chain reaction where one marker gets mismatched then every
        other marker gets mismatched too. In these cases you can set this parameter to True, and any double matched
        markers will simply be deleted.
    :type AllowDoubleMatching: bool, optional
    :param sorting_method: 'radial' or 'nearest'. This is only important if WarpSearchData is True; in that case,
        we are building a motion model on-the-fly based on previously seen markers, so the order we see them can be
        important. For data on a sphere, you may have more success by setting this to 'nearest', but this is a bit
        of an untested feature
    :type sorting_method: str, optional
    :param n_refernce_markers: the n inner_most Reference markers are used to align the ground truth to the MR volume
        before  matching. this can correct for setup errors. Note that we always move the reference. the best way
        for this to not matter either way is to not have any set up error!
    :type n_refernce_markers: int, optional
    """

    def __init__(self, GroundTruthData, DistortedData, reverse_gradient_data=None, WarpSearchData=True,
                 AutomatchMarkers=True, AllowDoubleMatching=False, sorting_method='radial', n_refernce_markers=0):

        # warping parameters:
        self.WarpSearchData = WarpSearchData
        self.AutomatchMarkers = AutomatchMarkers
        self._motion_estimate_update_rate = 1  # 1 = every found marker, which is potentially more accurate but slower.
        self.sorting_method = sorting_method
        self.AllowDoubleMatching = AllowDoubleMatching
        # match check parameters:
        self._mean_match_tolerance = 20  # warning raised if mean larger than this
        self._max_match_tolerance = 30  # warning raised if max larger than this

        self._n_reference_markers = n_refernce_markers

        # marker data:
        self.ground_truth_centroids = GroundTruthData.MarkerCentroids
        self.distorted_centroids = DistortedData.MarkerCentroids

        if reverse_gradient_data is None:
            self.distorted_centroidsRev = None
        else:
            self.distorted_centroidsRev = reverse_gradient_data.MarkerCentroids

        self._check_input_data()
        # run analysis:
        if self._n_reference_markers > 0:
            self._align_reference()

        if self.AutomatchMarkers:
            self.distorted_centroids = self._sort_distorted_centroids(self.distorted_centroids)
            self._CentroidMatch = self._match_distorted_markers_to_ground_truth(self.distorted_centroids)
            self._check_match()
            if self.distorted_centroidsRev is not None:
                self.distorted_centroidsRev = self._sort_distorted_centroids(self.distorted_centroidsRev)
                self._CentroidMatchRev = self._match_distorted_markers_to_ground_truth(self.distorted_centroidsRev)
        else:
            self._create_marker_data_prematched_markers()
        self._handle_double_matched_markers()

        # analyse marker positions
        self._generate_marker_position_data()

    def _check_input_data(self):
        """
        put any tests on the input data here to make sure the rest of the code can run.

        - tests that there at least as many ground truth centroids than distorted centroids
        - tests that data appears to be in mm (requirement)
        """
        if not self.ground_truth_centroids.shape[0] >= self.distorted_centroids.shape[0]:
            raise ValueError('There are fewer ground truth centroids than distorted centroids; this means the disorted'
                             'centroids can never be matched')

        if not self.AutomatchMarkers:
            if not self.ground_truth_centroids.shape[0] == self.distorted_centroids.shape[0]:
                raise ValueError(
                    f'Since you have set AutomatchMarkers=False, there must be equal numbers of ground truth'
                    f'markers and distorted markers, but ground truth has {self.ground_truth_centroids.shape[0]}'
                    f'and distorted has {self.distorted_centroids.shape[0]}.')

        if self.distorted_centroidsRev is not None:
            if not self.ground_truth_centroids.shape[0] >= self.distorted_centroidsRev.shape[0]:
                raise ValueError('There are fewer ground truth centroids than reversed gradient distorted centroids;'
                                 ' this means the disorted centroids can never be matched. Quitting')

        if np.max([self.ground_truth_centroids.x.max(), self.ground_truth_centroids.y.max(),
                   self.ground_truth_centroids.z.max()]) < 50:
            logger.warning('it appears that the ground truth centroids may not be defined in mm?'
                           'this will cause the algorithm to fail. Continuing but double check this!')

        if np.max([self.distorted_centroids.x.max(), self.distorted_centroids.y.max(),
                   self.distorted_centroids.z.max()]) < 50:
            logger.warning('it appears that the distorted centroids may not be defined in mm?'
                           'this will cause the algorithm to fail. Continuing but double check this!')
        if self.distorted_centroidsRev is not None:
            if np.max([self.distorted_centroidsRev.x.max(), self.distorted_centroidsRev.y.max(),
                       self.distorted_centroidsRev.z.max()]) < 50:
                logger.warning('it appears that the reverse gradient distorted centroids may not be defined in mm? '
                               'this will cause the algorithm to fail. Continuing but double check this!')

    def _get_grid_spacing(self):
        """
        Get the spacing between markers using the ground truth data. This is used to build the kernel for RBF
        interpolation during the marker matching step.
        """
        self._x_grid_spacing = np.mean(np.diff(np.sort(np.unique(self.ground_truth_centroids.x))))
        self._y_grid_spacing = np.mean(np.diff(np.sort(np.unique(self.ground_truth_centroids.y))))
        self._z_grid_spacing = np.mean(np.diff(np.sort(np.unique(self.ground_truth_centroids.z))))

    def _generate_extrapolant_models(self, CentroidsToSearch, distorted_centroids):
        """
        Use the found marker positions to extrapolate the distortion at other positions.

        :param CentroidsToSearch: Current search data. At this point this is based on the
            ground truth data with matched markers removed.

        returns X_dist, Y_dist, Z_dist, which is the distortion at every marker position in
        CentroidsToSearch.
        """
        points = self.ground_truth_centroids.iloc[self._MatchInd][['x', 'y', 'z']]
        distortion = self.ground_truth_centroids.iloc[self._MatchInd].to_numpy() - \
                     distorted_centroids.iloc[0:len(self._MatchInd)].to_numpy()

        # Nearest Neighbore interpolator
        self._X_interpolator = NearestNDInterpolator(points, distortion[:, 0])
        self._Y_interpolator = NearestNDInterpolator(points, distortion[:, 1])
        self._Z_interpolator = NearestNDInterpolator(points, distortion[:, 2])

        points = CentroidsToSearch[['x', 'y', 'z']].to_numpy().squeeze()
        X_dist = self._X_interpolator(points)
        Y_dist = self._Y_interpolator(points)
        Z_dist = self._Z_interpolator(points)
        return X_dist, Y_dist, Z_dist

    def _align_reference(self, rotation=True):
        """
        Moves the ground truth markers to be aligned with the distorted markers based on the central n refernce markers
        """
        if self.ground_truth_centroids is None or self.distorted_centroids is None:
            logger.warning('No centroids loaded')
            return

        def _calculate_radial_distance(centroids):
            """
            insert the radial value of each marker:
            """
            centroids['r'] = centroids.apply(
                lambda row: np.sqrt(row[0] ** 2 + row[1] ** 2 + row[2] ** 2), axis=1)
            if centroids.r.mean() < 10:
                logger.warning('it appears that your input data is in m, not mm - please use mm!')
            return centroids

        def _match_crosshair(search_centroids, reference_centroids):
            """
            match the distorted/ ground truth
            """

            _matched_reference = pd.DataFrame(columns=['x', 'y', 'z'])
            search_centroids.reset_index(inplace=True, drop=True)
            reference_centroids.reset_index(inplace=True, drop=True)
            for index, row in reference_centroids.iterrows():
                # Calculate distance from this marker to each item in search list
                Distances = cdist(search_centroids, np.atleast_2d(row), metric='euclidean')
                # Match marker with closest and save it
                ind_nearest = np.argmin(Distances)
                matched = search_centroids.iloc[ind_nearest]
                _matched_reference.loc[index] = matched

            return _matched_reference

        # Find the central position in the ground truth data
        ground_truth = self.ground_truth_centroids.copy()
        central_x = (ground_truth.x.min() + ground_truth.x.max()) / 2
        central_y = (ground_truth.y.min() + ground_truth.y.max()) / 2
        central_z = (ground_truth.z.min() + ground_truth.z.max()) / 2

        # Calculate the distance between each marker and central position
        ground_truth['r_centre'] = ground_truth.apply(
            lambda row: np.sqrt((row[0] - central_x) ** 2 + (row[1] - central_y) ** 2 + (row[2] - central_z) ** 2),
            axis=1)

        # Find the crosshair reference markers
        gt_reference = ground_truth.nsmallest(self._n_reference_markers, 'r_centre')[['x', 'y', 'z']]
        dist_reference = self.distorted_centroids.nsmallest(self._n_reference_markers, 'r')[['x', 'y', 'z']]

        # Calculate translation vector
        translation_vector = dist_reference.mean() - gt_reference.mean()

        # Calculate rotation vector
        if self._n_reference_markers < 3:
            rotation = False
        else:
            matched_reference = _match_crosshair(gt_reference + translation_vector, dist_reference)
            rotation_vector, error = transform.Rotation.align_vectors(dist_reference, matched_reference)

        # Transform centroids
        aligned = self.ground_truth_centroids[['x', 'y', 'z']] + translation_vector

        if rotation is True:
            aligned = rotation_vector.apply(aligned)

        aligned = pd.DataFrame(aligned, columns=['x', 'y', 'z'])
        self.ground_truth_centroids = _calculate_radial_distance(aligned)

    def _match_distorted_markers_to_ground_truth(self, CentroidsToMatch):
        """
        find the closest ground truth marker to a given distortion. The logic of this algorithm is based on Paul
        Liu's matlab version

        1. Distorted data is sorted by r so that the smallest (least distorted) markers are processed first
        2. the match is simply based on the minimum distance in cartesian space between the distorted marker and the
            search markers.
        3. Once a match has been found, that ground truth is removed from the search space
        4. if self.WarpSearchData=True, the search markers are warped according to the previously found data. For cases
            where distortion is small this is probably unnecessary, but for extreme distortions this algorithm will fail
            without this
        """

        self._MatchInd = []
        distances = []
        CentroidsToSearch = self.ground_truth_centroids.copy()
        for index, row in CentroidsToMatch.iterrows():

            XA = CentroidsToSearch[["x", "y", "z"]].to_numpy()
            XB = row[0:3].values
            # Calculate distance from this marker to each item in search list
            Distances = cdist(XA, np.atleast_2d(XB), metric='euclidean')
            # Match marker with closest and save it
            ind_closest_point_in_ground_truth = np.argmin(Distances)
            self._MatchInd.append(ind_closest_point_in_ground_truth)
            distances.append(np.linalg.norm(
                self.ground_truth_centroids.iloc[ind_closest_point_in_ground_truth][['x', 'y', 'z']] - row[
                    ['x', 'y', 'z']]))
            if not self.AllowDoubleMatching:
                CentroidsToSearch.iloc[ind_closest_point_in_ground_truth] = pd.Series(
                    {'x': 1000, 'y': 1000, 'z': 1000, 'r': 1732})

            # warp search data?
            if self.WarpSearchData and (index % self._motion_estimate_update_rate == 0) and index > 0:
                # get extrapolated motion at all points:
                X_dist, Y_dist, Z_dist = self._generate_extrapolant_models(CentroidsToSearch, CentroidsToMatch)
                # regenerate centroids to search:
                CentroidsToSearch = self.ground_truth_centroids.copy()
                # remove found entries:
                if not self.AllowDoubleMatching:
                    CentroidsToSearch.iloc[self._MatchInd] = pd.Series({'x': 1000, 'y': 1000, 'z': 1000, 'r': 1732})
                # update the data with the extrapolated motion
                CentroidsToSearch[['x']] = CentroidsToSearch[['x']] - X_dist[:, None]
                CentroidsToSearch[['y']] = CentroidsToSearch[['y']] - Y_dist[:, None]
                CentroidsToSearch[['z']] = CentroidsToSearch[['z']] - Z_dist[:, None]

        # create a new data frame that has the distorted and ground truth markers
        MatchedMarkers = self.ground_truth_centroids.iloc[self._MatchInd]
        MatchedMarkers = MatchedMarkers.rename(columns={"x": "x_gt", "y": "y_gt", "z": "z_gt", "r": "r_gt"})
        MatchedMarkers.reset_index(inplace=True, drop=True)
        MatchedCentroids = pd.concat([CentroidsToMatch, MatchedMarkers], axis=1)
        MatchedCentroids['match_distance'] = distances

        return MatchedCentroids

    def _check_match(self):
        """
        just a simple test to detect where bad matches have been made
        could add more sophisticated things in here if needed
        :return:
        """

        _throw_warning = False
        if self._CentroidMatch.match_distance.mean() > self._mean_match_tolerance:
            _throw_warning = True
        if self._CentroidMatch.match_distance.max() > self._max_match_tolerance:
            _throw_warning = True

        if _throw_warning:
            warn(f'\n\nThe marker match may have failed.\n\nThe mean detected distortion is {self._CentroidMatch.match_distance.mean(): 1.1f} mm'
                 f'and the max is {self._CentroidMatch.match_distance.max(): 1.1f}.'
                 f'\nYou can continue by pressing any key, but you should visualize the data using the plot_3D_markers'
                 f' method')
            input("Press any key to continue...")

    def _handle_double_matched_markers(self):
        """
        Checks if any distorted markers have been matched to the same ground truth point
        """
        non_unique_ind = self._CentroidMatch.duplicated(subset=['x_gt', 'y_gt', 'z_gt'], keep=False)
        if any(non_unique_ind):
            logger.error(f'{np.count_nonzero(non_unique_ind)} distorted centroids were matched to the same ground truth'
                         f'marker. This may be indicative of a serious issue and you should proceed with caution. For '
                         f'now, handling this by removing these entries from the analysis pool and continuing...')
            self._CentroidMatch.drop(self._CentroidMatch.index[non_unique_ind])
            self._CentroidMatch.reset_index(drop=True)

        if self.distorted_centroidsRev is not None:
            non_unique_ind = self._CentroidMatchRev.duplicated(subset=['x_gt', 'y_gt', 'z_gt'], keep=False)
            if any(non_unique_ind):
                logger.error(
                    f'{np.count_nonzero(non_unique_ind)} distorted centroids were matched to the same ground truth'
                    f'marker. This may be indicative of a serious issue and you should proceed with caution. For '
                    f'now, handling this by removing these entries from the analysis pool and continuing...')
                self._CentroidMatchRev.drop(self._CentroidMatchRev.index[non_unique_ind])
                self._CentroidMatchRev.reset_index(drop=True)

    def _create_marker_data_prematched_markers(self):
        """
        If no autosorting is requested, this creates the MatchedCentroids DataFrame directly from the input data
        with no sorting etc.
        """
        self._CentroidMatch = self.ground_truth_centroids
        self._CentroidMatch = self._CentroidMatch.rename(columns={"x": "x_gt", "y": "y_gt", "z": "z_gt", "r": "r_gt"})
        self._CentroidMatch = pd.concat([self._CentroidMatch, self.distorted_centroids], axis=1)

        if self.distorted_centroidsRev is not None:
            self._CentroidMatchRev = self.ground_truth_centroids
            self._CentroidMatchRev = self._CentroidMatchRev.rename(
                columns={"x": "x_gt", "y": "y_gt", "z": "z_gt", "r": "r_gt"})
            self._CentroidMatchRev = pd.concat([self._CentroidMatchRev, self.distorted_centroidsRev], axis=1)

    def _generate_marker_position_data(self):
        """
        Convert the marker position data to the data we want to use for calculating gradient and B0 distortion
        If only one dataset has been entered, then this is trivial. However, if we have a forward and reverse dataset
        then we need to use these two seperate the gradient and B0 effects.
        """
        if self.distorted_centroidsRev is None:
            # then our life is easy
            self.MatchedCentroids = self._CentroidMatch
            self.MatchedCentroids = self.MatchedCentroids.rename(
                columns={'x': 'x_gnl', 'y': 'y_gnl', 'z': 'z_gnl', 'r': 'r_gnl'})
        else:
            # then we have to do some work to seperate G and B0

            # first we will want to correlate the marker positions by sorting by ground truth r
            self._CentroidMatch = self._CentroidMatch.sort_values('r_gt').reset_index(drop=True)
            self._CentroidMatchRev = self._CentroidMatchRev.sort_values('r_gt').reset_index(drop=True)
            # now find an index where the ground truth markers match; we will discard other markers
            rev_match_index = []
            forward_match_index = []
            tol = 1e-1  # just a small number not a sensitve parameter
            n_unmatched_markers = 0
            for index, row in self._CentroidMatch.iterrows():
                # find the closest ground truth point in the reversed data
                XA = self._CentroidMatchRev[["x_gt", "y_gt", "z_gt"]].to_numpy()
                XB = row[['x_gt', 'y_gt', 'z_gt']].values
                # Calculate distance from this marker to each item in search list
                Distances = cdist(XA, np.atleast_2d(XB), metric='euclidean')
                # calculate location of minimum
                min_dist_ind = np.argmin(Distances)
                # if it is less than tolerance, record the position
                if Distances[min_dist_ind] < tol:
                    rev_match_index.append(min_dist_ind)
                    forward_match_index.append(index)
                else:
                    n_unmatched_markers += 1  # don't actually need this as can infer from data sizes.

            if n_unmatched_markers > 0:
                logger.warning(
                    f'failed to match {n_unmatched_markers} of a total {self.ground_truth_centroids.shape[0]}'
                    f' input markers. These will be deleted from returned data')
            # now we have indexes linking our two datasets based on their ground truth match
            self._CentroidMatchRev = self._CentroidMatchRev.iloc[rev_match_index].reset_index()
            self._CentroidMatch = self._CentroidMatch.iloc[forward_match_index].reset_index()

            # marker position distortion from gradients is difference between forward and reversed:
            self.MatchedCentroids = (self._CentroidMatch[['x', 'y', 'z', 'r']] + self._CentroidMatchRev[
                ['x', 'y', 'z', 'r']]) / 2
            self.MatchedCentroids = self.MatchedCentroids.rename(
                columns={'x': 'x_gnl', 'y': 'y_gnl', 'z': 'z_gnl', 'r': 'r_gnl'})
            # add ground truth to data frame
            self.MatchedCentroids[['x_gt', 'y_gt', 'z_gt', 'r_gt']] = self._CentroidMatch[
                ['x_gt', 'y_gt', 'z_gt', 'r_gt']]
            # marker position distortion from B0 is difference between from total and gradient distortion:
            '''
            deta_z_B0=z_dis_AP-z_GNL; %% maximum distortion is 16mm
            deta_y_B0=y_dis_AP-y_GNL;  %% maximum distortio is 4mm  
            '''
            B0_dis = self.MatchedCentroids[['x_gnl', 'y_gnl', 'z_gnl']].to_numpy() - \
                     self._CentroidMatch[['x', 'y', 'z']].to_numpy()
            B0_pos = B0_dis + self.MatchedCentroids[['x_gt', 'y_gt', 'z_gt']].to_numpy()
            self.MatchedCentroids[['x_B0', 'y_B0', 'z_B0']] = B0_dis

    def _sort_distorted_centroids(self, data_to_sort):
        """
        sorts the data frame.
        sorting is important, because the search for a matching ground truth marker is carried out in order
        """

        if self.sorting_method == 'radial':
            # sort input data by r
            data_to_sort = data_to_sort.sort_values('r')
            data_to_sort.reset_index(inplace=True, drop=True)
        if self.sorting_method == 'closest':
            # find the distorted marker that has the closest match to a ground truth marker.
            min_distance = []
            XA = self.ground_truth_centroids[["x", "y", "z"]].to_numpy()
            for index, row in data_to_sort.iterrows():
                XB = row[0:3].values
                Distances = cdist(XA, np.atleast_2d(XB), metric='euclidean')
                min_distance.append(np.min(Distances))
            starting_marker_ind = np.argmin(min_distance)
            # now calculate the distances of every other distorted marker to this marker:
            XA = data_to_sort[["x", "y", "z"]].to_numpy()
            XB = np.atleast_2d(data_to_sort.loc[starting_marker_ind][["x", "y", "z"]].to_numpy())
            Distances = cdist(XA, XB)
            distance_to_start_ind = np.argsort(Distances, axis=0)
            assert distance_to_start_ind[0] == starting_marker_ind
            data_to_sort = data_to_sort.reindex(np.squeeze(distance_to_start_ind))
            data_to_sort.reset_index(inplace=True, drop=True)

        return data_to_sort

    # public methods

    def plot_3D_markers(self, add_arrows=True, title='3D marker positions'):  # pragma: no cover
        """
        Works very similarly to the MarkerVolume version, but plots both sets of markers and adds arrows

        :param add_arrows: if True, arrows drawn between matches
        :type add_arrows: bool, optional
        :param title: plot title
        :type title: str, optional
        """

        fig = plt.figure()
        axs = fig.add_subplot(111, projection='3d')
        axs.scatter(self.MatchedCentroids.x_gt, self.MatchedCentroids.y_gt, self.MatchedCentroids.z_gt)
        axs.scatter(self.MatchedCentroids.x_gnl, self.MatchedCentroids.y_gnl, self.MatchedCentroids.z_gnl)
        if add_arrows:
            mag_color = np.sqrt((self.MatchedCentroids.x_gt - self.MatchedCentroids.x_gnl) ** 2
                                + (self.MatchedCentroids.y_gt - self.MatchedCentroids.y_gnl) ** 2
                                + (self.MatchedCentroids.z_gt - self.MatchedCentroids.z_gnl) ** 2)
            axs.quiver(self.MatchedCentroids.x_gnl, self.MatchedCentroids.y_gnl, self.MatchedCentroids.z_gnl,
                       self.MatchedCentroids.x_gt - self.MatchedCentroids.x_gnl,
                       self.MatchedCentroids.y_gt - self.MatchedCentroids.y_gnl,
                       self.MatchedCentroids.z_gt - self.MatchedCentroids.z_gnl,
                       edgecolors=("black",))

        axs.set_xlabel('X [mm]')
        axs.set_ylabel('Y [mm]')
        axs.set_zlabel('Z [mm]')
        axs.set_title(title)
        axs.set_box_aspect((np.ptp(self.MatchedCentroids.x_gt), np.ptp(self.MatchedCentroids.y_gt),
                            np.ptp(self.MatchedCentroids.z_gt)))
        plt.legend(['ground truth', 'distorted'])
        plt.show()

    def plot_compressed_markers(self, z_max=20, z_min=-20, add_arrows=True, title=None):  # pragma: no cover
        """
        compresses the 3D markers in the z plane, allowing a 2D visualisation.

        :param z_max: maximum z coordiante of markers to includes
        :type z_max: float, optional
        :param z_min: minimum z coordiante of markers to includes
        :type z_min: float, optional
        :param add_arrows: if True, arrows drawn between matches
        :type add_arrows: bool, optional
        :param title: plot title
        :type title: str, optional

        """

        def get_markers_as_function_of_z():
            """
            finds an index of markers within tolerence of z_pos
            """
            data_ind = np.logical_and(self.MatchedCentroids.z_gt > z_min, self.MatchedCentroids.z_gt < z_max)
            z_loc_data = self.MatchedCentroids[data_ind]
            return z_loc_data

        def plot_markers_inner(plot_data):
            """
            clears exising axes and plots plot_data
            """
            axs.clear()
            z_offset = (plot_data.z_gt - plot_data.z_gnl)
            axs.scatter(plot_data.x_gnl, plot_data.y_gnl, c='darkturquoise')  # c=z_offset,cmap="magma"
            axs.scatter(plot_data.x_gt, plot_data.y_gt, c='darkgoldenrod')
            # axs.legend(['Distorted', 'Ground Truth'])
            if add_arrows:
                axs.quiver(plot_data.x_gnl, plot_data.y_gnl,
                           plot_data.x_gt - plot_data.x_gnl,
                           plot_data.y_gt - plot_data.y_gnl,
                           angles="xy", units='xy', scale=1, width=1)
            axs.set_xlim([-150, 150])
            axs.set_ylim([-150, 150])
            axs.set_xlabel('x [mm]', fontsize=15)
            axs.set_ylabel('y [mm]', fontsize=15)
            axs.tick_params(axis='both', which='major', labelsize=12)
            axs.tick_params(axis='both', which='minor', labelsize=12)
            axs.axis("equal")
            axs.grid()
            plt.subplots_adjust(bottom=0.25)
            plt.legend(['distorted', 'ground truth'])
            if title:
                plt.title(title)

        fig, axs = plt.subplots(figsize=[8, 8], ncols=1, nrows=1)

        plot_data = get_markers_as_function_of_z()
        plot_markers_inner(plot_data)



import pathlib
import numpy as np
import logging
import json
from .utilities import get_dicom_data

ch = logging.StreamHandler()
formatter = logging.Formatter('[%(filename)s: line %(lineno)d %(levelname)8s] %(message)s')
ch.setFormatter(formatter)
logger = logging.getLogger(__name__)
logger.addHandler(ch)
logger.setLevel(logging.INFO)
logger.propagate = False


class ConvertMatchedMarkersToBz:
    """
    :param MatchedCentroids: a pandas data frame containing columns ['x_gt', 'y_gt', 'z_gt', 'x_gnl', 'y_gnl', 'z_gnl'].
        If you want to calculate B0, is should also have ['x_B0', 'y_B0', 'z_B0']. Such a dataframe is normally created
        within the MatchedMarkerVolumes class.
    :param dicom_data: either a dictionary or a path to a .json file. The minimume required parameters are below.
        dicom_data is calculated automatically when MR data is used to create a MarkerVolume, and can be exported
        to json using::

            marker_volume.save_dicom_data().

        dicom_data example::

            dicom_data = {'FOV': [400, 400, 400],
            'bandwidth': 203,
            'gama': 42.6,
            'image_size': [256, 256, 256]}
    :type dicom_data: dictionary or path

    """

    def __init__(self, MatchedCentroids, dicom_data):
        """
        init
        """

        self.dicom_data = get_dicom_data(dicom_data)
        self.MatchedCentroids = MatchedCentroids
        self.MagneticFields = self.MatchedCentroids[['x_gt', 'y_gt', 'z_gt']].copy()
        self.MagneticFields = self.MagneticFields .rename(columns={"x_gt": "x", "y_gt": "y", "z_gt": "z"})
        # ^ aim is to calculate B field at the ground truth marker positions
        if self.MatchedCentroids.r_gt.mean() < 10 or self.MatchedCentroids.r_gnl.mean() < 10:
            logger.warning('it appears that your input data is in m, not mm - please use mm!')
        self._check_dicom_data()
        self._calculate_Bz()

    def _check_dicom_data(self):
        """
        check that the minium data we will be needing is there.
        """
        expected_keys = ['FOV', 'bandwidth', 'gama', 'pixel_spacing', 'image_size', 'gradient_strength']
        for key in expected_keys:
            if not key in self.dicom_data.keys():
                raise AttributeError(f'missing required dicom_data field: {key}')

    def _get_B0_calc_direction(self):
        """
        if we have information about the frequency encode direction and information about B0 distortion,
        we should find that the max B0 distortion occurs in the frequency direction.
        If not, something problably is wrong.
        in any case, the direction of maximum B0 distortion is returned and used to calcualte B0
        """

        x_B0_mean = np.mean(np.abs(self.MatchedCentroids.x_B0))
        y_B0_mean = np.mean(np.abs(self.MatchedCentroids.y_B0))
        z_B0_mean = np.mean(np.abs(self.MatchedCentroids.z_B0))
        max = np.max([x_B0_mean, y_B0_mean, z_B0_mean])
        directions = ['x', 'y', 'z']
        direction = [int(el == max) for el in [x_B0_mean, y_B0_mean, z_B0_mean]]
        self._B0_direction_bool = [bool(el) for el in direction]
        if 'freq_encode_direction' in self.dicom_data.keys():
            # check that the maximum distortion matches the frequence encode direction
            if not directions[np.where(direction)[0][0]] == self.dicom_data['freq_encode_direction']:
                logger.warning(f'\nYou have entered a frequency code direction in '
                               f'{self.dicom_data["freq_encode_direction"]}, but maximum B0 distortion is detected'
                               f'in {directions[np.where(direction)[0][0]]}. Proceeding using '
                               f'{self.dicom_data["freq_encode_direction"]} but you should check this...\n')
            self._B0_direction_bool = [self.dicom_data["freq_encode_direction"]==el for el in directions]
        else:
            logger.warning("\nyou have input data on B0 distortion, but dicom_data['freq_encode_direction'] does not exist."
                           "Therefore, B0 fields will be calculated based on the direction of maximum B0 distortion, "
                           "which is \n")
        self._B0_direction_string = np.array(['x_B0', 'y_B0', 'z_B0'])[self._B0_direction_bool][0]

    def _calculate_Bz(self):
        """
        Calculate Gradient Bz at each point. This is based on code from Shanshan.
        Shanshans notes:
        - When reversing gradient polarity, the readout (z) and phase encoding (y) directions are reversed,
         however, the slice selction direction (x) is not.
        - B0 inhomogeneity distortion appears at readout diretion and is neglectable at phase encoding direction.
        - This is distorted marker position caused by GNL-only along readout and phase encoding directions
        """
        gradient_strength = np.array(self.dicom_data['gradient_strength'])  # unit(T / m)
        # ^ this is a vector [gx, gy, gz]
        self.MagneticFields['B_Gx'] = self.MatchedCentroids.x_gnl * gradient_strength[0]
        self.MagneticFields['B_Gy'] = self.MatchedCentroids.y_gnl * gradient_strength[1]
        self.MagneticFields['B_Gz'] = self.MatchedCentroids.z_gnl * gradient_strength[2]

        # if possible, add in B0 to data
        B0_fields = ['x_B0', 'y_B0', 'z_B0']
        B0_calc_possible = all([el in self.MatchedCentroids.columns for el in B0_fields])
        if B0_calc_possible:
            self._get_B0_calc_direction()
            self.MagneticFields['B0'] = self.MatchedCentroids[[self._B0_direction_string]] * \
                                        gradient_strength[self._B0_direction_bool][0] * 1e-3





"""
This package is for the characterisation and reporting of geometric distortion in MRI
Please see here for detailed documentation
https://acrf-image-x-institute.github.io/MRI_DistortionQA/index.html
"""

__version__ = '0.10.2'
from .calculate_harmonics import calculate_harmonics



import sys
import os
from pathlib import Path
import numpy as np
import logging
import plotly.express as px
import pandas as pd
from .utilities import get_gradient_spherical_harmonics, convert_cartesian_to_spherical, reconstruct_Bz
from .utilities import convert_spherical_to_cartesian
import plotly.graph_objects as go
from distutils.dir_util import copy_tree
from datetime import datetime
from jinja2 import Template
from .utilities import get_dicom_data

ch = logging.StreamHandler()
formatter = logging.Formatter('[%(filename)s: line %(lineno)d %(levelname)8s] %(message)s')
ch.setFormatter(formatter)
logger = logging.getLogger(__name__)
logger.addHandler(ch)
logger.setLevel(logging.INFO)  # This toggles all the logging in your app
logger.propagate = False


class TG284_Tests:  # pragma: no cover
    """
    tests defined here
    https://aapm.onlinelibrary.wiley.com/doi/full/10.1002/mp.14695
    """
    def test_distortion_less_than_1mm_within_r_100(self):
        """
        distortion test to be run by MRI_QA_Reporter
        """
        test_data = self._extract_data_from_MatchedMarkerVolume(r_max=100)
        if test_data.abs_dis.max() < 1:
            return True
        else:
            return False

    def test_distortion_less_than_2mm_within_r_200(self):
        """
        distortion test to be run by MRI_QA_Reporter
        """
        test_data = self._extract_data_from_MatchedMarkerVolume(r_max=200)
        if test_data.abs_dis.max() < 1:
            return True
        else:
            return False


class DefaultTestSuite:  # pragma: no cover
    """
    these are the tests which are run if no others are specified
    """

    def test_distortion_less_than_2mm_within_r_100(self):
        """
        distortion test to be run by MRI_QA_Reporter
        """
        test_data = self._extract_data_from_MatchedMarkerVolume(r_max=100)
        if test_data.abs_dis.max() < 2:
            return True
        else:
            return False

    def test_distortion_less_than_4mm_within_r_150(self):
        """
        distortion test to be run by MRI_QA_Reporter
        """
        test_data = self._extract_data_from_MatchedMarkerVolume(r_max=150)
        if test_data.abs_dis.max() < 4:
            return True
        else:
            return False


class Elekta_Distortion_tests:  # pragma: no cover
    """
    Geometric tests for Elekta system as described here:
    https://aapm.onlinelibrary.wiley.com/doi/epdf/10.1002/mp.14764
    """

    def test_distortion_less_than_1_mm_within_r_100_mm(self):
        """
        distortion test to be run by MRI_QA_Reporter
        """
        test_data = self._extract_data_from_MatchedMarkerVolume(r_max=100)
        if test_data.abs_dis.max() < 1:
            return True
        else:
            return False

    def test_distortion_less_than_2_mm_within_r_150_mm(self):
        """
        distortion test to be run by MRI_QA_Reporter
        """
        test_data = self._extract_data_from_MatchedMarkerVolume(r_max=150)
        if test_data.abs_dis.max() < 2:
            return True
        else:
            return False

    def test_distortion_less_than_4_mm_within_r_200_mm(self):
        """
        distortion test to be run by MRI_QA_Reporter
        """
        test_data = self._extract_data_from_MatchedMarkerVolume(r_max=200)
        if test_data.abs_dis.max() < 4:
            return True
        else:
            return False


class MRI_QA_Reporter:
    """
    generate a report based on either some matched data, or harmonics. In the latter case, report data is reconstructed.

    :param MatchedMarkerVolume: pandas dataframe containing  ['x_gt', 'y_gt', 'z_gt', 'x_gnl', 'y_gnl', 'z_gnl']
        You must enter one of MatchedMarkerVolume or gradient_harmonics
    :type MatchedMarkerVolume: pandas.DataFrame, optional
    :param gradient_harmonics: list of either three pandas series or three paths to csv files
        from SphericalHarmonicFit.
        List elements should correspond to [G_x_harmonics, G_y_harmonics, G_z_harmonics].
        You must enter one of MatchedMarkerVolume or gradient_harmonics
    :type gradient_harmonics: list, optional
    :param B0_harmonics: pandas series from SphericalHarmonicFit or path to saved csv file. If entered, is used
        to report and plot B0 homogeneity
    :type B0_harmonics: pandas.Series or path, optional
    :param recon_coords: The coordinates you want to calculate the fields at. a pandas Dataframe which can contain
        whatever you want, but MUST contain columns called: x, y, z. If left as None a default is generated.
    :type recon_coords: pandas.DataFrame, optional
    :param tests_to_run: a class containing any tests to run. The class in this instance serves only as a name
        space. All methods inside the class starting with 'test' are identified as tests. The test methods will
        have access to everything inside this class.
    :type tests_to_run: class, optional
    :param dicom_data: either a dictionary or path to a saved json file of the dicom_data object generated by
        MarkerVolume. This data is required for harmonic reconstruction. For MatchedMarkerVolume it is not required,
        but it does contain useful information about the acquisition so still a good idea to use it.
    :type dicom_data: dict or path to json, optional
    :param r_outer: for the harmonic approach, r_outer is used to determine where the data is reconstructed. It is
        also used to control the limits in which data is plotted
    :type r_outer: float, optional
    :param show_plots: if True, all plots will be shown in your browser. Otherwise they are only saved
    :type show_plots: bool, optional
    :param style: control the report style. At the moment can only be 'light' or 'dark'
    :type style: str
    """

    def __init__(self, MatchedMarkerVolume=None, gradient_harmonics=None, B0_harmonics=None,
                 recon_coords_cartesian=None, tests_to_run=DefaultTestSuite, dicom_data=None,
                 r_outer=None, show_plots=False, style='dark'):

        self._html_template_loc = Path(__file__).parent.resolve() / 'jinja_templates' / 'MR_report.html'
        self._show_plots = show_plots
        _available_plotly_themes = ["plotly", "plotly_white", "plotly_dark", "ggplot2",
                                    "seaborn", "simple_white", "none"]
        self._style = style
        self._set_styles()
        self._jinja_dict = {}  # this will hold all the info jinja needs to render
        self._test_class = tests_to_run
        self.r_outer = r_outer
        self.recon_coords_cartesian = recon_coords_cartesian
        self.dicom_data = get_dicom_data(dicom_data)
        self._build_acquisition_dict()
        self._get_analysis_data(MatchedMarkerVolume, gradient_harmonics, B0_harmonics)
        self._update_MatchedMarkerVolume()

        # put plots here
        self._plot_distortion_v_r()
        self._plot_3D_cutplanes()
        self._plot_B0_surface()
        self._build_homogeneity_report_table()
        # self._plot_gradients_surface()
        self._get_test_results()

    def _build_acquisition_dict(self):
        """
        will try and extract acquisition data from dicom_data, and put it in a new dictionary
        if this new dictionary exists, we will include the acquisition data in the report
        """
        if self.dicom_data is None:
            return
        self._jinja_dict['acquisition_data'] = {}
        _fields_of_interest = ['acquisition_date', 'magnetic_field_strength', 'bandwidth', 'pixel_spacing',
                               'freq_encode_direction', 'manufacturer', 'imaging_frequency']
        for key in self.dicom_data:
            if key in _fields_of_interest:
                new_key = key.replace('_', ' ')
                self._jinja_dict['acquisition_data'][new_key] = self.dicom_data[key]

    def _set_styles(self):
        """
        sets the report style by changing the plotly theme and chanding the html css file
        """
        if self._style == 'dark':
            self._plotly_theme = 'plotly_dark'
            self._html_theme = "themes/d42ker-github.css"
        elif self._style == 'light':
            self._plotly_theme = 'plotly_white'
            self._html_theme = "themes/whitey.css"
        else:
            raise AttributeError(f'unknown style entered: {self._style}. allowed options are "light" or "dark"')

    def _unique_name_generator(self, save_loc, input_name, rel_path=True):
        """
        take input name, append the current date to it
        then check if this new name already exists in save_loc; if it does add integers to it (_1, _2, etc.) until it doesn't
        """
        if not save_loc.is_dir():
            # in this case, the code will likely fail downstream but we don't have to do anything here
            return input_name
        input_name, extension = os.path.splitext(input_name)
        now = datetime.now()  # current date and time
        date_string = now.strftime("%d_%m_%Y")
        new_name = input_name + '_' + date_string
        appended_int = 0
        while (save_loc / (new_name + extension)).is_file():
            appended_int = appended_int + 1
            new_name = new_name + '_' + str(appended_int)
        new_name = new_name + extension
        if rel_path:
            new_name = (save_loc  / new_name).relative_to(save_loc.parent)
        else:
            new_name = save_loc / new_name
        return new_name

    def _get_test_results(self):
        """
        runs through all tests found in the _test_class and keeps all the results in _jinja_dict
        """

        self._jinja_dict['test_collection'] = {}
        self._jinja_dict['TestSuiteName'] = self._test_class.__name__
        method_list = [method for method in dir(self._test_class) if method.startswith('__') is False]
        # this will list all methods except 'dunder' methods e.g. __init__
        for test in method_list:
            if test[:4] == 'test':
                # then we found a true test
                test_name = test.replace('_', ' ') # under scores to space
                test_name = test_name.replace('test ','' )  # remove test from the name
                test_string = getattr(self._test_class, test)(self)
                if isinstance(test_string, bool):
                    if test_string:
                        test_string = "<span style='color:green'>Pass</span>"
                    else:
                        test_string = "<span style='color:red'>Fail</span>"
                elif not isinstance(test_string, str):
                    raise TypeError('please return test results as either booleans or strings')
                self._jinja_dict['test_collection'][test_name] = test_string

    def _get_analysis_data(self, MatchedMarkerVolume, gradient_harmonics, B0_harmonics):
        """
        check what the user has input, and attach to self
        if necessary, convert gradient_harmonics to distortion map
        """

        if MatchedMarkerVolume is not None and gradient_harmonics is not None:
            raise AttributeError(
                f'you cannot enter both MatchedMarkerVolume and gradient_harmonics at the same time...')

        if MatchedMarkerVolume is not None:
            self._check_marker_volume(MatchedMarkerVolume)
            self._MatchedMarkerVolume = MatchedMarkerVolume.copy()

        if gradient_harmonics is not None:
            self._check_dicom_data()
            self._Gx_Harmonics, self._Gy_Harmonics, self._Gz_Harmonics = \
                get_gradient_spherical_harmonics(gradient_harmonics[0], gradient_harmonics[1], gradient_harmonics[2])
            self._Gx_Harmonics = self._Gx_Harmonics * self.dicom_data['gradient_strength'][0]
            self._Gy_Harmonics = self._Gy_Harmonics * self.dicom_data['gradient_strength'][1]
            self._Gz_Harmonics = self._Gz_Harmonics * self.dicom_data['gradient_strength'][2]
            if self.r_outer is None:
                logger.warning('no r_outer value entered, using 150 mm')
                self.r_outer = 150
            if self.recon_coords_cartesian is None:
                self.recon_coords_cartesian = self._generate_recon_coords()
            self.recon_coords = convert_cartesian_to_spherical(self.recon_coords_cartesian)
            self._reconstruct_gradient_fields()
            self._convert_magnetic_field_to_distortion()

        if B0_harmonics is not None:
            if isinstance(B0_harmonics, pd.Series):
                self.B0_harmonics = B0_harmonics
            elif isinstance(B0_harmonics, (str, Path)):
                self.B0_harmonics = pd.read_csv(B0_harmonics, index_col=0).squeeze("columns")
            else:
                raise AttributeError('could not read in B0_harmonics...please input either a series or a '
                                     'path to a saved csv file')
        else:
            self.B0_harmonics = None

    def _convert_magnetic_field_to_distortion(self):
        """
        converts reconstructed gradient fields into geometric distortion.
        Essentially this process is doing the opposite of FieldCalculation.ConvertMatchedMarkersToBz
        """
        self._MatchedMarkerVolume = pd.DataFrame()
        self._MatchedMarkerVolume = self._MatchedMarkerVolume.assign(x_gt=self.recon_coords_cartesian.x)
        self._MatchedMarkerVolume = self._MatchedMarkerVolume.assign(y_gt=self.recon_coords_cartesian.y)
        self._MatchedMarkerVolume = self._MatchedMarkerVolume.assign(z_gt=self.recon_coords_cartesian.z)
        bandwidth = np.array(self.dicom_data['bandwidth'])
        image_size = np.array(self.dicom_data['image_size'])
        gama = np.array(self.dicom_data['gama'])
        FOV = np.array(self.dicom_data['FOV'])

        gradient_strength = bandwidth * image_size / (gama * 1e6 * FOV * 1e-3)  # unit(T / m)
        # ^ this is a vector [gx, gy, gz]
        self._MatchedMarkerVolume = \
            self._MatchedMarkerVolume.assign(x_gnl=self.Gx_Bfield / (gradient_strength[0]))
        self._MatchedMarkerVolume = \
            self._MatchedMarkerVolume.assign(y_gnl=self.Gy_Bfield / (gradient_strength[1]))
        self._MatchedMarkerVolume = \
            self._MatchedMarkerVolume.assign(z_gnl=self.Gz_Bfield / (gradient_strength[2]))

    def _reconstruct_gradient_fields(self):
        """
        utilise the utilities.reconstruct_Bz function to reconsutrct each of the gradient fields at recon_coords
        """
        self.Gx_Bfield = reconstruct_Bz(self._Gx_Harmonics, self.recon_coords, quantity='T')
        self.Gy_Bfield = reconstruct_Bz(self._Gy_Harmonics, self.recon_coords, quantity='T')
        self.Gz_Bfield = reconstruct_Bz(self._Gz_Harmonics, self.recon_coords, quantity='T')

    def _build_homogeneity_report_table(self):
        """
        Generate a table of predicted pk-pk homogeneity at different r
        :return:
        """
        if self.B0_harmonics is None:
            return
        radii_of_interest = [50, 100, 150, 200]
        pk_pk = []
        for DSV_radius in radii_of_interest:
            azimuth = np.linspace(0, 2 * np.pi, 100)
            elevation = np.linspace(0, np.pi, 100)
            [AZ, EL, R] = np.meshgrid(azimuth, elevation, DSV_radius, indexing='ij')
            surface_coords = pd.DataFrame({'r': R.flatten(), 'azimuth': AZ.flatten(), 'elevation': EL.flatten()})
            surface_coords = convert_spherical_to_cartesian(surface_coords)
            B0 = reconstruct_Bz(self.B0_harmonics, surface_coords, quantity='T', r_outer=DSV_radius)
            pk_pk.append(int((B0.max() - B0.min()) * 1e6))
        _B0_stats = pd.DataFrame({'r': radii_of_interest, 'pk_pk [\u03BCT]': pk_pk} )

        self._B0_table = go.Figure(data=[go.Table(
            header=dict(values=list(_B0_stats.columns),
                        align='left'),
            cells=dict(values=[_B0_stats.r, _B0_stats['pk_pk [\u03BCT]']],
                       align='left'))
        ])
        self._B0_table.update_layout(template=self._plotly_theme)

    def _check_dicom_data(self):
        """
        check that if input dicom data is required, it confirms to minimum standard
        """
        _required_fields = ['FOV', 'bandwidth', 'gama', 'image_size', 'gradient_strength']

        if not isinstance(self.dicom_data, dict) or not all(
                name in self.dicom_data.keys() for name in _required_fields):
            raise AttributeError(f'dicom_data must be a dict with at least the following keys: {_required_fields}')

    def _check_marker_volume(self, MatchedMarkerVolume):
        """
        check that if a MatchedMarkerVolume is entered, is meets expected format
        """
        _required_cols = ['x_gt', 'y_gt', 'z_gt', 'x_gnl', 'y_gnl', 'z_gnl']
        if not isinstance(MatchedMarkerVolume, pd.DataFrame) or \
                not all(name in MatchedMarkerVolume.columns for name in _required_cols):
            raise AttributeError(f'MatchedMarkerVolume must be a dataframe containing columns'
                                 f'[x_gt, y_gt, z_gt, x_gnl, y_gnl, z_gnl]')

    def _generate_recon_coords(self):
        """
        Generate a grid of coordinates to perform reconstruction
        """
        x = np.linspace(-self.r_outer, self.r_outer, 50)
        y = np.linspace(-self.r_outer, self.r_outer, 50)
        z = np.linspace(-self.r_outer, self.r_outer, 50)
        [x_recon, y_recon, z_recon] = np.meshgrid(x, y, z, indexing='ij')
        recon_coords = pd.DataFrame({'x': x_recon.flatten(), 'y': y_recon.flatten(), 'z': z_recon.flatten()})
        return recon_coords

    def _plot_distortion_v_r(self):
        """
        generate a histogram as a function of r
        """

        # data over different planes
        xy_plot_data = self._extract_data_from_MatchedMarkerVolume(r_max=self.r_outer, z_select=0)
        xy_plot_data['plane'] = 'XY'
        zx_plot_data = self._extract_data_from_MatchedMarkerVolume(r_max=self.r_outer, y_select=0)
        zx_plot_data['plane'] = 'ZX'
        zy_plot_data = self._extract_data_from_MatchedMarkerVolume(r_max=self.r_outer, x_select=0)
        zy_plot_data['plane'] = 'ZY'
        # create a new data frame combined from these for easy plotting
        all_dsv_plot_data = self._extract_data_from_MatchedMarkerVolume(r_max=self.r_outer)
        all_dsv_plot_data['plane'] = 'Entire DSV'


        plot_data = pd.concat([all_dsv_plot_data, xy_plot_data, zx_plot_data, zy_plot_data], ignore_index=True, axis=0)
        self._fig_distortion_v_r = px.scatter(plot_data, x="r_gt", y="abs_dis",color="plane",
                                              labels={
                                                  "r_gt": "Distance from the center (mm)", #rename the x axis
                                                  "abs_dis": " Absolute Distortion (mm)", #rename the y axis
                                              },
                                              title=f"Cardinal plane data, {self.r_outer} DSV")


        self._fig_distortion_v_r.update_layout(template=self._plotly_theme)
        if self._show_plots:
            self._fig_distortion_v_r.show()

    def _plot_3D_cutplanes(self):
        """
        create a 3D cone plot in each of the three cardinal directions
        """
        xy_plot_data = self._extract_data_from_MatchedMarkerVolume(r_max=self.r_outer, z_select=0)
        zx_plot_data = self._extract_data_from_MatchedMarkerVolume(r_max=self.r_outer, y_select=0)
        zy_plot_data = self._extract_data_from_MatchedMarkerVolume(r_max=self.r_outer, x_select=0)
        plot_data = pd.concat([xy_plot_data, zx_plot_data, zy_plot_data], ignore_index=True, axis=0)

        self._fig_3D_planes = go.Figure(data=go.Cone(
            x=plot_data['x_gt'],
            y=plot_data['y_gt'],
            z=plot_data['z_gt'],
            u=plot_data['x_dis'],
            v=plot_data['y_dis'],
            w=plot_data['z_dis'],
            colorbar=dict(title= 'mm'),
            sizemode="absolute",
            sizeref=10))


        self._fig_3D_planes.update_layout(scene=dict(aspectratio=dict(x=1, y=1, z=0.8),
                                                     camera_eye=dict(x=1.2, y=1.2, z=0.6)))
        self._fig_3D_planes.update_layout(template=self._plotly_theme, scene = dict(
                    xaxis_title='X Axis (mm)',
                    yaxis_title='Y Axis (mm)',
                    zaxis_title='Z Axis (mm)'))
        if self._show_plots:
            self._fig_3D_planes.show()

    def _plot_B0_surface(self):
        """
        if self.B0_harmonics exists, reconstruct B0 on the surface of r_outer and make a pretty 3D plot
        """

        if self.B0_harmonics is None:
            return
        # generate surface recon coordinates
        if self.r_outer is not None:
            r = self.r_outer
        else:
            r = 150
        azimuth = np.linspace(0, 2 * np.pi, 100)
        elevation = np.linspace(0, np.pi, 100)
        [AZ, EL, R] = np.meshgrid(azimuth, elevation, r, indexing='ij')
        surface_coords = pd.DataFrame({'r': R.flatten(), 'azimuth': AZ.flatten(), 'elevation': EL.flatten()})
        surface_coords = convert_spherical_to_cartesian(surface_coords)

        B0 = reconstruct_Bz(self.B0_harmonics, surface_coords, quantity='uT', r_outer=r)
        x = surface_coords.x.to_numpy().reshape(100, 100)
        y = surface_coords.y.to_numpy().reshape(100, 100)
        z = surface_coords.z.to_numpy().reshape(100, 100)
        B0 = B0.to_numpy().reshape(100, 100)

        self._fig_DSV_surface = go.Figure(data=[go.Surface(z=z, x=x, y=y,colorbar=dict(title= 'B(uT)'), surfacecolor=B0)])

        self._fig_DSV_surface.update_layout(title='DSV surface [uT]', autosize=True,
                          template=self._plotly_theme, scene = dict(
                    xaxis_title='X Axis (mm)',
                    yaxis_title='Y Axis (mm)',
                    zaxis_title='Z Axis (mm)') )

        if self._show_plots:
            self._fig_DSV_surface.show()

    def _update_MatchedMarkerVolume(self):
        """
        adds in 'r_gt', 'r_gnl' and 'dis_x,y,z' to _MatchedMarkerVolume input
        """
        self._MatchedMarkerVolume['r_gt'] = np.sqrt(self._MatchedMarkerVolume.x_gt ** 2
                                                    + self._MatchedMarkerVolume.y_gt ** 2
                                                    + self._MatchedMarkerVolume.z_gt ** 2)
        self._MatchedMarkerVolume['r_gnl'] = np.sqrt(self._MatchedMarkerVolume.x_gnl ** 2
                                                     + self._MatchedMarkerVolume.y_gnl ** 2
                                                     + self._MatchedMarkerVolume.z_gnl ** 2)
        self._MatchedMarkerVolume['x_dis'] = self._MatchedMarkerVolume.x_gnl - self._MatchedMarkerVolume.x_gt
        self._MatchedMarkerVolume['y_dis'] = self._MatchedMarkerVolume.y_gnl - self._MatchedMarkerVolume.y_gt
        self._MatchedMarkerVolume['z_dis'] = self._MatchedMarkerVolume.z_gnl - self._MatchedMarkerVolume.z_gt
        self._MatchedMarkerVolume['abs_dis'] = np.sqrt(self._MatchedMarkerVolume.x_dis ** 2
                                                       + self._MatchedMarkerVolume.y_dis ** 2
                                                       + self._MatchedMarkerVolume.z_dis ** 2)
        # get resolution in each direction
        self._x_res = np.mean(np.diff(np.sort(self._MatchedMarkerVolume.x_gt.unique())))
        self._y_res = np.mean(np.diff(np.sort(self._MatchedMarkerVolume.y_gt.unique())))
        self._z_res = np.mean(np.diff(np.sort(self._MatchedMarkerVolume.z_gt.unique())))

    def _extract_data_from_MatchedMarkerVolume(self, x_select=None, y_select=None, z_select=None,
                                               max_num_points=None, r_max=None):
        """
        Extract data from the matched volume for use in direct plotting

        uses the values of _x_res etc. to define a tolerance
        """
        if (x_select is None) and (y_select is None) and (z_select is None) and (r_max is None):
            logger.warning('you should set a value for at least one of x_ind, y_ind, and z_ind or r_max')
        # start with everything selected, overwrite below
        x_ind = np.ones(self._MatchedMarkerVolume.shape[0])
        y_ind = np.ones(self._MatchedMarkerVolume.shape[0])
        z_ind = np.ones(self._MatchedMarkerVolume.shape[0])
        r_ind = np.ones(self._MatchedMarkerVolume.shape[0])

        if x_select is not None:
            x_ind = np.logical_and(self._MatchedMarkerVolume.x_gt <= (x_select + self._x_res / 2),
                                   self._MatchedMarkerVolume.x_gt >= (x_select - self._x_res / 2))
        if y_select is not None:
            y_ind = np.logical_and(self._MatchedMarkerVolume.y_gt <= (y_select + self._y_res / 2),
                                   self._MatchedMarkerVolume.y_gt >= (y_select - self._y_res / 2))
        if z_select is not None:
            z_ind = np.logical_and(self._MatchedMarkerVolume.z_gt <= (z_select + self._z_res / 2),
                                   self._MatchedMarkerVolume.z_gt >= (z_select - self._z_res / 2))
        if r_max is not None:
            r_ind = self._MatchedMarkerVolume.r_gt <= r_max

        select_ind = np.logical_and(np.logical_and(x_ind, y_ind), z_ind)
        select_ind = np.logical_and(select_ind, r_ind)
        plot_data = self._MatchedMarkerVolume[select_ind]
        plot_data = plot_data.reset_index(drop=True)
        if (max_num_points is not None) and plot_data.shape[0] > max_num_points:
            every_nth_point = int(plot_data.shape[0] / max_num_points)
            plot_data = plot_data[plot_data.reset_index().index % every_nth_point == 0]

        return plot_data

    def _set_up_directory_structure(self):
        """
        ensure we have a place to save the reports.
        will set up a directory in
        """
        if not os.path.isdir(self.output_folder):
            os.makedirs(self.output_folder)
        if not os.path.isdir(self.output_folder / 'plots'):
            os.mkdir(self.output_folder / 'plots')

        # copy themes
        theme_dir = (Path(__file__).parent / 'jinja_templates' / 'themes').resolve()
        copy_tree(str(theme_dir), str(self.output_folder / 'themes'))

    def _get_template(self):
        """
        loads the jinja template
        """

        with open(self._html_template_loc) as f:
            template_text = f.read()
        j2_template = Template(template_text)
        return j2_template

    # public methods

    def write_html_report(self, output_folder=None):
        """
        Generates a html report encompassing available acquisition information, test results, and intercative
        plotly figures

        :param output_folder: folder to write report. Defaults to ~/Documents/MR_QA_Reports
        :type output_folder: Path or string, optional
        """
        if output_folder is None:
            self.output_folder = Path(os.path.expanduser('~')) / 'Documents' / 'MR_QA_Reports'
        else:
            self.output_folder = Path(output_folder)
        self._set_up_directory_structure()

        # save plots and update jinja_dict

        distortion_v_r_save_name = self._unique_name_generator(self.output_folder / 'plots', 'distortion_v_r.html')
        self._fig_distortion_v_r.write_html(self.output_folder / distortion_v_r_save_name, full_html=False, include_plotlyjs='cdn')
        self._jinja_dict['dist_v_r_source'] = distortion_v_r_save_name

        threeD_plane_save_name = self._unique_name_generator(self.output_folder / 'plots', '3D_planes.html')
        self._fig_3D_planes.write_html(self.output_folder / threeD_plane_save_name, full_html=False, include_plotlyjs='cdn')
        self._jinja_dict['cutplanes_source'] = threeD_plane_save_name

        if self.B0_harmonics is not None:
            DSV_surface_save_name = self._unique_name_generator(self.output_folder / 'plots', 'DSV_surface.html',)
            self._fig_DSV_surface.write_html(self.output_folder / DSV_surface_save_name, full_html=False, include_plotlyjs='cdn')
            self._jinja_dict['dsv_surf_source'] = DSV_surface_save_name

            B0_table_save_name = self._unique_name_generator(self.output_folder / 'plots', 'B0_table.html')
            self._B0_table.write_html(self.output_folder / B0_table_save_name, full_html=False, include_plotlyjs='cdn')
            self._jinja_dict['B0_table_source'] = B0_table_save_name
        else:
            self._jinja_dict['dsv_surf_source'] = None

        # set up template
        self._jinja_dict['html_theme_loc'] = self._html_theme
        j2_template = self._get_template()
        report_name = self._unique_name_generator(self.output_folder, 'MR_QA_report.html', rel_path=False)
        report_code = os.path.split(report_name)[1]  # for printing
        report_code = os.path.splitext(report_code)[0]
        report_code = report_code.replace('MR_QA_report_','')
        self._jinja_dict['report_name'] = report_code
        report_string = j2_template.render(self._jinja_dict)
        with open(report_name, 'w') as f:
            f.write(report_string)

        print(f'The report has been compiled and can be found at {self.output_folder}')



"""
General purpose reusable functions
"""

import os
import pathlib
import sys
import glob
import logging
from pathlib import Path
import numpy as np
from scipy.special import lpmv
import pandas as pd
from matplotlib import pyplot as plt
import pydicom
import json


ch = logging.StreamHandler()
formatter = logging.Formatter('[%(filename)s: line %(lineno)d %(levelname)8s] %(message)s')
ch.setFormatter(formatter)
logger = logging.getLogger(__name__)
logger.addHandler(ch)
logger.setLevel(logging.INFO)  # This toggles all the logging in your app
logger.propagate = False


class bcolors:
    """
    This is just here to enable me to print pretty colors to the linux terminal
    """
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKCYAN = '\033[96m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'


def build_dicom_affine(Dicomfiles):
    """
    Build the dicom affine matrix.
    The reference I used to create the affine is `NIbabel <https://nipy.org/nibabel/dicom/dicom_orientation.html>`_
    (we are using the r/c defined under 'DICOM affines again')

    :param Dicomfiles: list of pydicom.dataset.FileDataset instances (created using pydicom.read_file)
    :type Dicomfiles: list
    :return: CoordinateMatrix, a 4x4 matrix which transforms the image based coordinates (i,j,k) to world coordinates (x,y,z)
    """

    assert isinstance(Dicomfiles, list)
    assert isinstance(Dicomfiles[0], pydicom.dataset.FileDataset)

    ds = Dicomfiles[0]  # just use first one, they all give the same result
    if not np.logical_or(np.isclose(np.abs(ds.ImageOrientationPatient), 1),
                         np.isclose(np.abs(ds.ImageOrientationPatient), 0)).all():
        '''
        This just checks that the ImageOrientationPatient tag only contains 0 or 1, i.e. that no weird oblique
        orientation is in use. For now I don't want to support or think about non cardinal encoding directions
        '''
        raise NotImplementedError('at the moment we only support cardinal imaging planes')

    CoordinateMatrix = np.zeros([4, 4])
    F = np.array([ds.ImageOrientationPatient[3:], ds.ImageOrientationPatient[:3]]).T
    F = np.round(F)
    first_col = np.multiply(F[:, 0], ds.PixelSpacing[0])
    sec_col = np.multiply(F[:, 1], ds.PixelSpacing[1])
    CoordinateMatrix[:, 0][0:3] = first_col
    CoordinateMatrix[:, 1][0:3] = sec_col
    CoordinateMatrix[:, 3][0:3] = [float(i) for i in ds.ImagePositionPatient]
    if Dicomfiles.__len__() > 1:
        # figure out slice direction
        SliceLoc = []
        image_position_patient_list = []
        for file in Dicomfiles:
            # need to extract the slice coordinates one at a time (can't think of a better way anyway...)
            image_position_patient_list.append(file.ImagePositionPatient)
            SliceLoc.append(file.SliceLocation)
        SliceLoc = np.array(SliceLoc)
        image_position_patient_list = np.array(image_position_patient_list)
        if not np.allclose(np.diff(SliceLoc), np.diff(SliceLoc)[0], atol=.1, rtol=.1):
            logger.error(f'it appears there may be missing or inconsistently spaced slices. Continuing')
        k = (image_position_patient_list[-1] - image_position_patient_list[0])/(image_position_patient_list.shape[0]-1)
        CoordinateMatrix[0:3, 2] = k

    return CoordinateMatrix


def printProgressBar (iteration, total, prefix = '', suffix = '', decimals = 1, length = 100, fill = '█'):
    """
    [credit here](https://gist.github.com/greenstick/b23e475d2bfdc3a82e34eaa1f6781ee4)

    Call in a loop to create terminal progress bar

        iteration   - Required  :
        total       - Required  :
        prefix      - Optional  :
        suffix      - Optional  :
        decimals    - Optional  :
        length      - Optional  :
        fill        - Optional  :
        printEnd    - Optional  : end character (e.g. "\r", "\r\n") (Str)

    :param iteration: current iteration (Int)
    :param total: total iterations (Int)
    :param prefix:  prefix string (Str)
    :param suffix: suffix string (Str)
    :param decimals: positive number of decimals in percent complete (Int)
    :param length: character length of bar (Int)
    :param fill: bar fill character (Str)
    :return: progress_string: the string to print
    """
    percent = ("{0:." + str(decimals) + "f}").format(100 * (iteration / float(total)))
    filledLength = int(length * iteration // total)
    bar = fill * filledLength + '-' * (length - filledLength)
    progress_string = f'\r{prefix} |{bar}| {percent}% {suffix}'
    return progress_string


def dicom_to_numpy(path_to_dicoms, FilesToReadIn=None, file_extension='dcm', return_XYZ=False,
                   zero_pad=0, enforce_increasing_coords=False):
    """
    This function does two things:

    1. Loads all the dicom files in path_to_dicoms in a numpy array. data is ordered according to [row, column, slice]
    2. Builds the matrix that transforms the [r,c,s] indices into [x,y,z] coordinates

    :param path_to_dicoms: folder where all dicoms images are stored
    :type path_to_dicoms: pathlib.Path or string
    :param FilesToReadIn: By default, this function will read in all files of file type file_extension. If you only want
        to read in a subset of files, you can specify them with FilesToReadIn
    :type FilesToReadIn: list, optional
    :param file_extension: extension of dicom files, e.g. 'dcm' or 'IMA'
    :type file_extension: string
    :param return_XYZ: if True, will build and return three matrics containing the X, Y, and Z coords at each i,j,k
        pixel.
    :type return_XYZ: bool, optional
    :param zero_pad: this many zeros will be placed around the returned volume and the coordinates updated accordingly.
        main use case was for distortion correction. Note that all returned objects will be self consistently affected
        by this parameter
    :type zero_pad: int, optional
    :returns: ImageArray: a numpy array of voxels
    :returns: dicom_affine: a matrix that transforms the [r,c,s] indices into [x,y,z] coordinates
    :returns: (X, Y, Z): Coordinate arrays the same size as ImageArray. Optional
    """
    if not FilesToReadIn:
        FilesToReadIn = get_all_files(path_to_dicoms, file_extension=file_extension)
    if not FilesToReadIn:
        raise FileNotFoundError(f'could not find any {file_extension} files in input directory {path_to_dicoms}'
                                f'\nTry using the ImExtenstion= input option to look for different image types')
    if isinstance(FilesToReadIn, list):
        CompletePathFiles = [str(Path(path_to_dicoms) / file) for file in FilesToReadIn]
    elif isinstance(FilesToReadIn, (str, pathlib.Path)):
        CompletePathFiles = [str(Path(path_to_dicoms) / FilesToReadIn)]

    dicom_slices = [pydicom.read_file(f) for f in CompletePathFiles]
    dicom_slices = sort_dicom_slices(dicom_slices)[0]
    dicom_affine = build_dicom_affine(dicom_slices)
    dicom_affine[0:3, 3] = dicom_affine[0:3, 3] + (-1*zero_pad*dicom_affine[0:3, 0:3].sum(axis=1))  # update start point for zero padding
    n_rows = dicom_slices[0].Rows + zero_pad*2
    n_cols = dicom_slices[0].Columns + zero_pad*2
    n_slices = len(dicom_slices) + zero_pad*2

    ImageArray = np.zeros([n_rows, n_cols, n_slices])
    if enforce_increasing_coords:
        # this option exists for the distortion correction code, which increasing indices
        new_affine = dicom_affine.copy()
        coordinates_end = dicom_affine @ [n_rows-1, n_cols-1, n_slices-1, 1]
        for col in [0, 1]:  # check first two cols
            for i, pixel_increment in enumerate(dicom_affine[:, col]):
                if pixel_increment < 0:
                    '''
                    then we need to force this to positive, and switch the start position to the 
                    current end position
                    '''
                    new_affine[i, col] = -1 * pixel_increment
                    new_start = coordinates_end[i]
                    new_affine[i, 3] = new_start
        dicom_affine = new_affine
    for i, file in enumerate(dicom_slices):
        if zero_pad > 0:
            ImageArray[zero_pad:-zero_pad, zero_pad:-zero_pad, i+zero_pad] = file.pixel_array
        else:
            ImageArray[:, :, i] = file.pixel_array
    if return_XYZ:
        rcsMatrix = np.zeros([4, n_rows * n_cols * n_slices])
        r_indices = np.arange(0, n_rows)
        c_indices = np.arange(0, n_cols)
        s_indices = np.arange(0, n_slices)
        [RR, CC, SS] = np.meshgrid(r_indices, c_indices, s_indices, indexing='ij')
        rcsMatrix[0, :] = RR.flatten()
        rcsMatrix[1, :] = CC.flatten()
        rcsMatrix[2, :] = SS.flatten()
        rcsMatrix[3, :] = 1
        XYZtemp = np.matmul(dicom_affine, rcsMatrix)
        X = np.reshape(XYZtemp[0, :], ImageArray.shape)
        Y = np.reshape(XYZtemp[1, :], ImageArray.shape)
        Z = np.reshape(XYZtemp[2, :], ImageArray.shape)

        assert ImageArray.shape == X.shape

        return ImageArray, dicom_affine, (X, Y, Z)
    else:
        return ImageArray, dicom_affine


def sort_dicom_slices(dicom_datasets):
    """
    sort slices by instance number
    :param dicom_datasets: list of pydicom files
    :return: sorted list of pydicom files
    """
    instance_number = [el.InstanceNumber for el in dicom_datasets]
    sort_ind = np.argsort(instance_number)
    return list(np.array(dicom_datasets)[sort_ind]), sort_ind


def get_all_files(PathToData, file_extension):
    """
    quick script to just collect all the files in the Analysis path

    :param PathToData: folder where the files are
    :type PathToData: pathlib.Path or string
    :param file_extension: extension of files to return, e.g. 'dcm'
    :type file_extension: str
    :returns Files: list of all found files
    """

    if not file_extension[0] == '.':
        # handles the case where the user entered 'dcm' instead of '.dcm'
        file_extension = '.' + file_extension
    file_extension = '*' + file_extension
    # check that this is now in the format we require
    if not file_extension[0:2] == '*.':
        logger.error('please enter the file_extension parameter like this : file_extension = "jpg"')
        sys.exit(1)

    if not isinstance(PathToData,Path):
        PathToData = Path(PathToData)

    if not os.path.isdir(PathToData):
        raise FileNotFoundError(f'invalid path supplied; {PathToData} does not exist')
    AllFiles = glob.glob(str(PathToData / file_extension))
    Files = []
    for file in AllFiles:
        head, tail = os.path.split(file)
        Files.append(tail)
    if not Files:
        logging.error(f'no {file_extension} files in {PathToData}')

    return Files


def convert_cartesian_to_spherical(InputCoords):
    """
    Converts cartesian coordinates [x,y,z] to spherical coordinates [r, azimuth, elevation].
    If r, azimuth, or elevation already exist as column names they are overwritten.
    The convention used here matches the convention
    `matlab uses <https://www.mathworks.com/help/matlab/ref/cart2sph.html>`_, except that we add pi to the azimuth and
    pi/2 to the elevation such that

    - 0 <= azimuth <= 2*pi
    - 0 <= elevation <= pi

    There was once a very good reason for doing this; I can't remember what it was anymore but I like to stick
    to tradition.

    :param InputCoords: a data frame containing (at least) columns x, y, z.
    :returns InputCoords: same data frame with additional columns [r, azimuth, elevation]
    """
    if 'r' in InputCoords.columns:
        InputCoords = InputCoords.drop(columns='r')
    if 'azimuth' in InputCoords.columns:
        InputCoords = InputCoords.drop(columns='azimuth')
    if 'elevation' in InputCoords.columns:
        InputCoords = InputCoords.drop(columns='elevation')

    InputCoords.insert(3, "r", np.sqrt(InputCoords.x ** 2 + InputCoords.y ** 2 + InputCoords.z ** 2))
    InputCoords.insert(4, "azimuth", np.arctan2(InputCoords.y, InputCoords.x) + np.pi)
    InputCoords.insert(5, "elevation", np.arctan2(InputCoords.z, np.sqrt(InputCoords.x ** 2 + InputCoords.y ** 2)) + np.pi/2)


    return InputCoords


def convert_spherical_to_cartesian(InputCoords):
    """
    Converts spherical coordinates [r, azimuth, elevation] to cartesian coordinates [x,y,z].
    if the cartesian columns 'x', 'y', 'z' already exist, they will be overwritten

    :param InputCoords: a data frame containing columns r, azimuth, elevation
    :returns InputCoords:   same data frame with additional columns [x,y,z]
    """
    if 'x' in InputCoords.columns:
        InputCoords = InputCoords.drop(columns='x')
    if 'y' in InputCoords.columns:
        InputCoords = InputCoords.drop(columns='y')
    if 'z' in InputCoords.columns:
        InputCoords = InputCoords.drop(columns='z')

    # InputCoords.insert(3, "x", InputCoords.r * np.sin(InputCoords.azimuth - np.pi) * np.cos(InputCoords.elevation - np.pi/2))
    # InputCoords.insert(4, "y", InputCoords.r * np.sin(InputCoords.azimuth - np.pi) * np.sin(InputCoords.elevation - np.pi/2))
    # InputCoords.insert(5, "z", InputCoords.r * np.cos(InputCoords.azimuth))

    InputCoords.insert(3, "x", InputCoords.r * np.cos(InputCoords.azimuth - np.pi) * np.cos(InputCoords.elevation - np.pi/2))
    InputCoords.insert(4, "y", InputCoords.r * np.sin(InputCoords.azimuth - np.pi) * np.cos(InputCoords.elevation - np.pi/2))
    InputCoords.insert(5, "z", InputCoords.r * np.sin(InputCoords.elevation - np.pi/2))

    return InputCoords


def generate_legendre_basis(InputCoords, n_order):
    """
    Produce spherical base at all coordinate values.
    This utilises the 'fully normalised' convention described here <https://au.mathworks.com/help/matlab/ref/legendre.html>`_
    Note that we scale all r values by 1000, which is to help ensure numerical stability. This is based on the assumption
    that this input data is in mm. The code will throw a warning if this does not appear to the case.

    :param InputCoords: a data frame which can contain whatever you want, but MUST contain columns called
        elevation, azimuth, and r
    :type InputCoords: Dataframe
    :param n_order: order of harmonic fit
    :type n_order: int
    :returns legendre_basis: a pandas data frame of size [Ncoords, (n_order +1)**2)]
    """

    if InputCoords.r.mean() < 10:
        logger.warning('it appears that your input data is in m, not mm - please use mm!')

    legendre_basis = np.zeros([len(InputCoords.index), (n_order+1)**2])
    k = 0
    coeff_names = []
    for n in range(0, n_order+1):  # the plus 1 is because range stops at -1 for some reason
        Norm = np.sqrt(n+0.5)
        P = Norm * lpmv(0, n, np.cos(InputCoords.elevation))
        legendre_basis[:, k] = np.multiply((InputCoords.r/1000) ** n, P)  # m=0 terms
        coeff_names.append(f'A_{n}_0')
        k = k + 1
        for j in range(0, n):
            m = j+1  # this looks weird but it works to give values m=1:n
            Norm = ((-1)**m) * np.sqrt( (n+0.5)*(np.math.factorial(n-m)) / np.math.factorial(n+m))
            P = Norm * lpmv(m, n, np.cos(InputCoords.elevation))
            legendre_basis[:,k] = np.multiply(np.multiply((InputCoords.r/1000) ** n, P), np.cos(m * InputCoords.azimuth))
            coeff_names.append(f'A_{n}_{m}')
            k = k + 1
            legendre_basis[:, k] = np.multiply(np.multiply((InputCoords.r/1000) ** n, P), np.sin(m * InputCoords.azimuth))
            coeff_names.append(f'B_{n}_{m}')
            k = k + 1

    return pd.DataFrame(legendre_basis, columns=coeff_names)


def convert_spherical_harmonics(harmonics, input_format='full', output_format='none'):
    """
    function to convert between different harmonic formalisms.
    At the moment only works between 'fully normalised' and 'unnormalised' but we could add more later if
    we need to.
    example::

        g_x_harmonics_no_norm = convert_spherical_harmonics(g_x_harmonics, input_format='full', output_format='none')

    :param harmonics: a pandas series or numpy array of size (n_order+1)^2, where the harmonics are ordered as
        A00, A10, A11, B11, A20, A21, B21, A22, B22...etc.
    :param input_format: format of input harmonic normalisation; 'none' or 'full'
    :param output_format: format of output harmonic normalisation; 'none' or 'full'
    :return: converted harmonics as a pandas series
    """

    formats = ['none', 'full']
    if input_format not in formats:
        raise NotImplementedError(f'no method exists to convert {input_format} format')
    if output_format not in formats:
        raise NotImplementedError(f'no method exists to convert {output_format} format')

    case_string = input_format.lower() + '_' + output_format.lower()
    if input_format.lower == output_format.lower():
        case_string = 'same'  # will do nothing, except return the data as a series

    converted_harmonics = harmonics.copy()
    n_order = np.sqrt(converted_harmonics.shape[0]) - 1
    assert n_order.is_integer()  # if this is false then an invalid number of harmonics is entered
    n_order = int(n_order)
    # generate normalisation factors:
    k = 0
    coeff_names = []
    Norm = []
    for n in range(0, n_order + 1):  # the plus 1 is because range stops at -1 for some reason
        Norm.append(np.sqrt(n + 0.5))
        coeff_names.append(f'A_{n}_0')
        k = k + 1
        for j in range(0, n):
            m = j + 1  # this looks weird but it works to give values m=1:n
            Norm.append(((-1) ** m) * np.sqrt((n + 0.5) * (np.math.factorial(n - m)) / np.math.factorial(n + m)))
            coeff_names.append(f'A_{n}_{m}')
            coeff_names.append(f'B_{n}_{m}')
            Norm.append(((-1) ** m) * np.sqrt((n + 0.5) * (np.math.factorial(n - m)) / np.math.factorial(n + m)))

    if case_string == 'full_none':
        converted_harmonics = np.multiply(harmonics, Norm)
    if case_string == 'none_full':
        converted_harmonics = np.divide(harmonics, Norm)
    if isinstance(converted_harmonics, np.ndarray):
        converted_harmonics = pd.DataFrame(converted_harmonics, index=coeff_names).squeeze()

    return converted_harmonics


def get_gradient_spherical_harmonics(Gx_Harmonics, Gy_Harmonics, Gz_Harmonics):
    """
    return the gradient spherical harmonics as as pandas series. this function is simply a clean way to handle
    the different ways users can specify harmonics to other componets of this code

    :param Harmonics:

    :param Gx_Harmonics: either a pandas series or a path to a csv. If it is already a series we do nothing, but handle
        this option for cleaner code elsewhere. Gy_Harmonics and Gz_Harmonics are the same
    :returns: Gx_Harmonics, Gy_Harmonics, Gz_Harmonics as Pandase series
    """

    if isinstance(Gx_Harmonics, pd.Series):
        Gx_Harmonics = Gx_Harmonics
    elif isinstance(Gx_Harmonics, (str,Path)):
        Gx_Harmonics = pd.read_csv(Gx_Harmonics, index_col=0).squeeze("columns")
    else:
        raise AttributeError('could not read in Gx harmonics...please input either a series or a '
                     'path to a saved csv file')


    if isinstance(Gy_Harmonics, pd.Series):
        Gy_Harmonics = Gy_Harmonics
    elif isinstance(Gy_Harmonics, (str,Path)):
        Gy_Harmonics = pd.read_csv(Gy_Harmonics, index_col=0).squeeze("columns")
    else:
        raise AttributeError('could not read in Gy harmonics...please input either a series or a '
                     'path to a saved csv file')


    if isinstance(Gz_Harmonics, pd.Series):
        Gz_Harmonics = Gz_Harmonics
    elif isinstance(Gz_Harmonics, (str,Path)):
        Gz_Harmonics = pd.read_csv(Gz_Harmonics, index_col=0).squeeze("columns")
    else:
        raise AttributeError('could not read in Gz harmonics...please input either a series or a '
                     'path to a saved csv file')

    return Gx_Harmonics, Gy_Harmonics, Gz_Harmonics


def get_dicom_data(dicom_data):
    """
    figures out whether dicom data is a dict or a path to a json file
    if the latter, reads into  a dict and returns
    """

    if isinstance(dicom_data, dict):
        return dicom_data
    elif isinstance(dicom_data,(pathlib.Path, str)):
        with open(dicom_data, 'r') as f:
            dicom_data = json.load(f)
        return dicom_data
    else:
        raise AttributeError(f'could not read in dicom_data of type {type(dicom_data)}')


def reconstruct_Bz(harmonics, coords, quantity='uT', r_outer=None):
    """
    Reconstruct Bz at an arbitrary set of coordinates.
    n_order of reconstruction is derived from the shape of harmonics

    :param harmonics: a pandas series of size (n_order+1)^2, such as that defined within
        FieldAnalysis.SphericalHarmonicFit
    :param coords: a pandas Dataframe which can contain whatever you want, but MUST contain columns called:
        elevation, azimuth, and r
    :param quantity:
        - 'T': return raw field in T
        - 'uT': return field in uT, with A00 set to 0.
    :param r_outer: optional parameter; if you try to reconstruct points with r > r_outer a warning is thrown
    :type r_outer: float or None
    :returns Bz_recon: Pandase series with the value of Bz at each element in coords
    """

    if not quantity in ['uT', 'T']:
        logger.error(f'unsupported quantity "{quantity}" entered for recon.'
                       f'\n valid options are "field" or "uT". Using uT')
        quantity = 'uT'

    if any(abs(coords.r) > r_outer):
        logger.warning(f'you are trying to reconstruct {np.count_nonzero(abs(coords.r) > r_outer)} '
                       f'data points outside rOuter. You can do it, but the futher away'
                       'from rOuter you get the less you should trust this.')

    n_order = int(np.sqrt(harmonics.shape[0])-1)
    legendre_recon = generate_legendre_basis(coords, n_order)
    if quantity == 'uT':  # field in uT, with A00 set to 0 (e.g. only looking at perturbation)
        harmonics_temp = harmonics.copy()
        harmonics_temp.A_0_0 = 0
        Bz_recon = legendre_recon @ harmonics_temp
        Bz_recon = Bz_recon * 1e6
    elif quantity == 'T':  # raw field in T
        Bz_recon = legendre_recon @ harmonics
    else:
        raise NotImplementedError(f'unknown reconstruction option entered: {quantity}')

    return Bz_recon


def compare_recon_report_with_ground_truth_report(ground_truth_report, recon_report, tolerance=1):  # pragma: no cover
    """
    This is to compare the distortion data contained within two Reports.MRI_QA_recon_reporter
    objects. The original purpose was to compare a recon_report generated from direct data, and one
    generated via harmonics, although in principle it can take any two reports.

    :param ground_truth_report: Reports.MRI_QA_recon_reporter instance
    :param recon_report: Reports.MRI_QA_recon_reporter instance
    :param tolerance: tolerance in mm
    """
    # direct overlay of real versus recon
    fig, axs = plt.subplots(nrows=2, ncols=2, figsize=[10, 10])
    if ground_truth_report.r_outer == recon_report.r_outer:
        # then use r_outer to filter the data
        gt_data = ground_truth_report._extract_data_from_MatchedMarkerVolume(r_max=ground_truth_report.r_outer)
        recon_data = recon_report._extract_data_from_MatchedMarkerVolume(r_max=recon_report.r_outer)
    else:
        gt_data = ground_truth_report._extract_data_from_MatchedMarkerVolume()
        recon_data = recon_report._extract_data_from_MatchedMarkerVolume()

    axs[0, 0].scatter(gt_data.r_gt, gt_data.abs_dis)
    axs[0, 0].scatter(recon_data.r_gt, recon_data.abs_dis)
    axs[0, 0].legend(['original_data', 'recon_data'])
    axs[0, 0].set_xlabel('r [mm]')
    axs[0, 0].set_ylabel('absolute distortion [mm]')

    # scatter of real versus recon
    axs[0, 1].scatter(gt_data.abs_dis, recon_data.abs_dis)
    axs[0, 1].set_xlabel('original data')
    axs[0, 1].set_ylabel('recon data')

    # box plot of errors
    error = gt_data.abs_dis - recon_data.abs_dis
    axs[1, 0].boxplot(error)
    axs[1, 0].set_ylabel('recon error [mm]')


    # get index of outliers
    outliers_ind = error > tolerance
    print(f'{np.count_nonzero(outliers_ind)} outliers exist, at index {np.where(outliers_ind)}')
    outliers_ind = np.where(outliers_ind)
    for ind in outliers_ind[0]:
        print(f'index {ind} is at {gt_data.x_gt.iloc[ind]: 1.1f},'
              f' {gt_data.y_gt.iloc[ind]: 1.1f},'
              f' {gt_data.z_gt.iloc[ind]: 1.1f}  the value in the ground truth is'
              f' {gt_data.abs_dis.iloc[ind]: 1.1f}'
              f' and the value in the reconstructed is {recon_data.abs_dis.iloc[ind]: 1.1f}')


def enumerate_subfolders(data_loc):
    """
    A simple function that prints all the subfolders in data_loc as a dict, e.g.

    {'1': 'folder1',
     '2': 'folder2'} etc

    I find this is a useful to create a dict object that I copy to the start of my analysis scripts
    """
    data_dict = {}
    subfolders = [ f.path for f in os.scandir(data_loc) if f.is_dir() ]
    for i, dir in enumerate(subfolders):
        folder_name = os.path.split(dir)[1]
        data_dict[str(i)] = folder_name
    for key in data_dict.keys():
        print(f"'{key}': '{data_dict[key]}',")


def plot_MarkerVolume_overlay(MarkerVolumeList, legend=None):  # pragma: no cover
    """
    Plot overlaid 3D scatter plots of the marker positions in each MarkerVolume

    :param MarkerVolumeList: list of MarkerVolume instances
    :param legend: legend to display on plot
    :return: None
    """

    assert isinstance(MarkerVolumeList, list)

    fig = plt.figure()
    axs = fig.add_subplot(111, projection='3d')
    for MarkerVolume in MarkerVolumeList:
        axs.scatter(MarkerVolume.MarkerCentroids.x, MarkerVolume.MarkerCentroids.y, MarkerVolume.MarkerCentroids.z)
        axs.set_xlabel('X [mm]')
        axs.set_ylabel('Y [mm]')
        axs.set_zlabel('Z [mm]')
        axs.set_title('3D marker positions')
    axs.set_box_aspect((np.ptp(MarkerVolume.MarkerCentroids.x),
                        np.ptp(MarkerVolume.MarkerCentroids.y),
                        np.ptp(MarkerVolume.MarkerCentroids.z)))
    if legend:
        plt.legend(legend)
    plt.show()


def plot_compressed_MarkerVolumes(MarkerVolumeList, z_max=20, z_min=-20, title=None, legend=None): # pragma: no cover
    """
    plot overlay of compressed marker volumes (z coord is ignored)

    :param MarkerVolumeList:  list of MarkerVolume instances
    :type MarkerVolumeList: list
    :param z_max: max z data to include
    :type z_max: float, optional
    :param z_min: min z data to include
    :type z_min: float, optional
    :param title: title of plot
    :type title: str, optional
    :param legend: legend of plot
    :type legend: list, optional
    :return: None
    """

    fig, axs = plt.subplots(figsize=[8, 8], ncols=1, nrows=1)
    for MarkerVolume in MarkerVolumeList:
        data_ind = np.logical_and(MarkerVolume.MarkerCentroids.z > z_min, MarkerVolume.MarkerCentroids.z < z_max)
        plot_data = MarkerVolume.MarkerCentroids[data_ind]
        axs.scatter(plot_data.x, plot_data.y)
        axs.set_xlim([-150, 150])
        axs.set_ylim([-150, 150])
        axs.set_xlabel('x [mm]', fontsize=15)
        axs.set_ylabel('y [mm]', fontsize=15)
        axs.tick_params(axis='both', which='major', labelsize=12)
        axs.tick_params(axis='both', which='minor', labelsize=12)
        axs.axis("equal")
        axs.grid()
        plt.subplots_adjust(bottom=0.25)
        if legend:
            plt.legend(legend)
        if title:
            plt.title(title)


def plot_matched_volume_hist(VolumeList, legend=None):
    """
    create a histogram of total distortion for all input volumes

    :param VolumeList: a list of [matched marker volumes](https://acrf-image-x-institute.github.io/MRI_DistortionQA/code_docs.html#MRI_DistortionQA.MarkerAnalysis.MatchedMarkerVolumes)
    :type VolumeList: list
    :param legend: legend to print
    :type legend: list, optional
    :return: None
    """
    bins = np.linspace(0, 10, 30)
    plt.figure()
    for marker_volume in VolumeList:
        plt.hist(marker_volume.MatchedCentroids.match_distance, bins=bins, alpha=0.5)
    if legend:
        plt.legend(['original', 'corrected'])
    plt.xlabel('distortion [mm]')
    plt.tight_layout()
    plt.show()


def plot_distortion_xyz_hist(MatchedMarkerVolume):
    """
    plot overlaid x, y, z distortion for an input instance of
    [MatchedMarkerVolume](https://acrf-image-x-institute.github.io/MRI_DistortionQA/code_docs.html#MRI_DistortionQA.MarkerAnalysis.MatchedMarkerVolumes)

    :param MatchedMarkerVolume: an instance of MatchedMarkerVolumes
    """

    x_dis = MatchedMarkerVolume.MatchedCentroids.x_gt - MatchedMarkerVolume.MatchedCentroids.x_gnl
    y_dis = MatchedMarkerVolume.MatchedCentroids.y_gt - MatchedMarkerVolume.MatchedCentroids.y_gnl
    z_dis = MatchedMarkerVolume.MatchedCentroids.z_gt - MatchedMarkerVolume.MatchedCentroids.z_gnl

    plt.figure()
    bins = np.linspace(0, 10, 30)
    plt.hist(abs(x_dis), bins=bins, alpha=0.5)
    plt.hist(abs(y_dis), bins=bins, alpha=0.5)
    plt.hist(abs(z_dis), bins=bins, alpha=0.5)
    plt.legend(['x', 'y', 'z'])
    plt.xlabel('distortion [mm]')
    plt.show()


def print_dict(dict):
    """
    just provides a nicer display of the input dict than the default print function,
    got sick of writing the same code over and over...
    """
    for key in dict.keys():
        print(f'{key}: {dict[key]}')




import os
import logging
import sys
import warnings
import pydicom
import matplotlib.pyplot as plt
from finufft import Plan
from finufft import nufft2d1, nufft2d2
import numpy as np
from scipy.fft import fft2
from scipy.fft import fftshift
from scipy.sparse.linalg import lsqr
from scipy.sparse.linalg import LinearOperator
from scipy.interpolate import RectBivariateSpline
import pandas as pd
from pathlib import Path
from .utilities import get_all_files, convert_cartesian_to_spherical, generate_legendre_basis, dicom_to_numpy
from .utilities import sort_dicom_slices
from .utilities import get_gradient_spherical_harmonics
from .utilities import printProgressBar
from time import perf_counter
from abc import abstractmethod

logging.basicConfig(format='[%(filename)s: line %(lineno)d] %(message)s', level=logging.WARNING)
logger = logging.getLogger(__name__)


class DistortionCorrectorBase:
    """
    This algorithm is based on `this work`_<https://onlinelibrary.wiley.com/doi/epdf/10.1002/mrm.25487>
    """

    def __init__(self, ImageDirectory, NufftLibrary='finufft', gradient_harmonics=None,
                 ImExtension='.dcm', dicom_data=None, correct_through_plane=True, pad=0):
        """
        :param ImageDirectory:
        :param NufftLibrary:
        :param Gx_Harmonics:
        :param Gy_Harmonics:
        :param Gz_Harmonics:
        :param ImExtension:
        :param dicom_data:
        :param correct_through_plane:
        """

        if 'DistortionCorrectorBase' in str(self.__class__):
            raise TypeError('DistortionCorrectorBase should not be called directly; it only exists for other correctors'
                         'to inherit. Quitting')

        self.correct_through_plane = correct_through_plane
        self._n_zero_pad = pad  # n_pixels to add around each edge of volume. set to 0 for no zero padding
        self._dicom_data = dicom_data
        self._Gx_Harmonics, self._Gy_Harmonics, self._Gz_Harmonics = \
            get_gradient_spherical_harmonics(gradient_harmonics[0], gradient_harmonics[1], gradient_harmonics[2])
        self._Gx_Harmonics = self._Gx_Harmonics * 1
        self._Gy_Harmonics = self._Gy_Harmonics * 1
        self._Gz_Harmonics = self._Gz_Harmonics * -1

        self.ImageDirectory = Path(ImageDirectory)
        self._all_dicom_files = get_all_files(self.ImageDirectory, ImExtension)
        CompletePathFiles = [self.ImageDirectory / file for file in self._all_dicom_files]
        dicom_slices = [pydicom.read_file(f) for f in CompletePathFiles]
        sort_ind = sort_dicom_slices(dicom_slices)[1]
        self._all_dicom_files = list(np.array(self._all_dicom_files)[sort_ind])
        self._n_dicom_files = len(self._all_dicom_files)


        self.ImageArray,\
        self._dicom_affine,\
        (self._X, self._Y, self._Z) = dicom_to_numpy(self.ImageDirectory,
                                                     file_extension=ImExtension,
                                                     return_XYZ=True,
                                                     zero_pad=self._n_zero_pad,
                                                     enforce_increasing_coords=False)
        self._get_rows_and_cols()
        self.n_order = int(np.sqrt(self._Gx_Harmonics.size) - 1)
        self.Images = get_all_files(self.ImageDirectory, ImExtension)
        self.r_DSV = 150  # only for drawing on the plot
        # Select nufft algorithm to use:
        if not NufftLibrary in ['pynufft', 'finufft', 'torchnufft']:
            raise NotImplementedError(f'Unknown NUFFTlibrary entered: {NufftLibrary}.'
                         f'\nAllowed options are: pynufft, finufft, torchnufft')
        self.NUFFTlibrary = NufftLibrary  # pynufft, finufft, or torchnufft
        self._check_input_data()

    @abstractmethod
    def _correct_image(self):
        """
        This is the method to correct a 2D slice of data; each inheriting corrector must supply its own _correct_image
        method.
        """
        pass

    def _get_rows_and_cols(self):
        """
        just extract the number of rows and columns from a dicom header - we need it for zero padding later.
        I suspect I should be able to derive this from the dicom affine, but this way seems more foolproof
        """
        demo_header = pydicom.read_file(self.ImageDirectory / self._all_dicom_files[0])
        self._Rows, self._Cols = self.ImageArray.shape[0:2]
        self._ImageOrientationPatient = demo_header.ImageOrientationPatient
        self._PixelSpacing = self._dicom_data['pixel_spacing']

    def _check_input_data(self):
        """
        Put any tests of input data here
        """
        assert self._Gx_Harmonics.shape == self._Gy_Harmonics.shape == self._Gz_Harmonics.shape
        assert self._n_zero_pad >= 0

    def _unpad_image_arrays(self):
        """
        remove any zero padding from original and reconstructed image volumes
        """
        if self._n_zero_pad > 0:
            self.ImageArray = self.ImageArray[self._n_zero_pad:-self._n_zero_pad,
                              self._n_zero_pad:-self._n_zero_pad,
                              self._n_zero_pad:-self._n_zero_pad]
            self._image_array_corrected = self._image_array_corrected[self._n_zero_pad:-self._n_zero_pad,
                                          self._n_zero_pad:-self._n_zero_pad,
                                          self._n_zero_pad:-self._n_zero_pad]

    def _calculate_encoding_fields(self):
        """
        Based on the spherical harmonics and the coordinates derived from the dicom, estimate the encoding
        fields that have been applied to each voxel
        """
        legendre_basis = generate_legendre_basis(self.coords, self.n_order)
        self.Gx_encode = (legendre_basis @ self._Gx_Harmonics)
        self.Gy_encode = (legendre_basis @ self._Gy_Harmonics)
        self.Gz_encode = (legendre_basis @ self._Gz_Harmonics)

    def _zero_volume(self, volume_to_zero):
        """
        takes an input volume and replaces all values <0 with 0.
        required for the image domain corrector which tends to produce -ve pixels
        """
        negative_ind = volume_to_zero < 0
        logger.warning(f'{np.count_nonzero(negative_ind)} negative pixels detected; setting these to zero and continuing')
        volume_to_zero[negative_ind] = 0
        return volume_to_zero

    def _generate_Kspace_data(self):
        """
        Get the k space of the current slice data.
        :return:
        """
        self.k_space = fftshift(fft2(fftshift(self._image_to_correct)))

    def _calculate_encoding_indices(self):
        """
        generates both the linear (i.e. assumed) indices and the distorted indices.
        These indices are passed to the NUFFT library.
        """

        x_lin_size, y_lin_size = self._image_to_correct.shape
        xn_lin = np.linspace(-x_lin_size / 2, -x_lin_size / 2 + x_lin_size - 1, x_lin_size)
        yn_lin = np.linspace(-y_lin_size / 2, -y_lin_size / 2 + y_lin_size - 1, y_lin_size)
        [xn_lin, yn_lin] = np.meshgrid(xn_lin, yn_lin, indexing='ij')
        xn_lin = xn_lin.flatten()
        yn_lin = yn_lin.flatten()
        self.sk = xn_lin / x_lin_size
        self.tk = yn_lin / y_lin_size

        if (np.round(self._ImageOrientationPatient) == [0, 1, 0, 0, 0, -1]).all():
            xn_dis = self.Gz_encode / (self._PixelSpacing[2])
            self.xj = xn_dis * 2 * np.pi
            yn_dis = self.Gy_encode / (self._PixelSpacing[1])
            self.yj = yn_dis * 2 * np.pi
        elif (np.round(self._ImageOrientationPatient) == [1, 0, 0, 0, 0, -1]).all():
            xn_dis = self.Gz_encode / (self._PixelSpacing[2])
            self.xj = xn_dis * 2 * np.pi
            yn_dis = self.Gx_encode / (self._PixelSpacing[0])
            self.yj = yn_dis * 2 * np.pi
        elif (np.round(self._ImageOrientationPatient) == [1, 0, 0, 0, 1, 0]).all():
            xn_dis = self.Gy_encode / (self._PixelSpacing[1])
            self.xj = xn_dis * 2 * np.pi
            yn_dis = self.Gx_encode / (self._PixelSpacing[0])
            self.yj = yn_dis * 2 * np.pi
        elif np.round(self._ImageOrientationPatient == [2, 2, 2, 2, 2, 2]).all():
            # this is for through plane correction where the real images are [0, 1, 0, 0, 0, -1]
            self.xj = pd.Series(xn_lin * 2 * np.pi)
            yn_dis = self.Gx_encode / (self._PixelSpacing[0])
            self.yj = pd.Series(yn_dis * 2 * np.pi)
            xn_dis = pd.Series(np.copy(xn_lin))
        elif np.round(self._ImageOrientationPatient == [3, 3, 3, 3, 3, 3]).all():
            # this is for through plane correction where the real images are [1, 0, 0, 0, 0, -1]
            self.xj = pd.Series(xn_lin * 2 * np.pi)
            yn_dis = self.Gy_encode / (self._PixelSpacing[1])
            self.yj = yn_dis * 2 * np.pi
            xn_dis = pd.Series(np.copy(xn_lin))
        elif (np.round(self._ImageOrientationPatient) == [1, 1, 1, 1, 1, 1]).all():
            # this is for through plane correction where the real images are [1, 0, 0, 0, 1, 0]
            self.xj = pd.Series(xn_lin * 2 * np.pi)
            yn_dis = -1*self.Gz_encode / (self._PixelSpacing[2])
            self.yj = yn_dis * 2 * np.pi
            xn_dis = pd.Series(np.copy(xn_lin))
        else:
            raise NotImplementedError('this slice orientation is not handled yet sorry')

        self.xj = self.xj.to_numpy()
        self.yj = self.yj.to_numpy()
        self.xn_dis_pixel = np.reshape((xn_dis - xn_lin).values, self._image_to_correct.shape)
        self.yn_dis_pixel = np.reshape((yn_dis - yn_lin).values, self._image_to_correct.shape)

    def _plot_encoding_fields(self, vmin=None, vmax=None):  # pragma: no cover
        """
        debug code; make sure the gradient fields look correct
        """
        fig, axs = plt.subplots(nrows=1, ncols=3, figsize=[15, 5])
        gah = np.squeeze(np.reshape(self.Gx_encode.to_numpy(), self._image_shape)) * 1e6

        Xim = axs[0].imshow(gah, vmin=vmin, vmax=vmax)
        axs[0].set_title('Gx')
        # axs[0].set_xlabel(self._row_label)
        # axs[0].set_ylabel(self._col_label)

        draw_circle = plt.Circle((0, 0), 150, fill=False)
        axs[0].add_artist(draw_circle)
        axs[0].grid(False)

        gah = np.squeeze(np.reshape(self.Gy_encode.to_numpy(), self._image_shape)) * 1e6

        Yim = axs[1].imshow(gah, vmin=vmin, vmax=vmax)
        axs[1].set_title('Gy')
        # axs[1].set_xlabel(self._row_label)
        # axs[1].set_ylabel(self._col_label)

        draw_circle = plt.Circle((0, 0), 150, fill=False)
        axs[1].add_artist(draw_circle)
        axs[1].grid(False)

        gah = np.squeeze(np.reshape(self.Gz_encode.to_numpy(), self._image_shape)) * 1e6

        Zim = axs[2].imshow(gah, vmin=vmin, vmax=vmax)
        axs[2].set_title('Gz')
        # axs[2].set_xlabel(self._row_label)
        # axs[2].set_ylabel(self._col_label)

        draw_circle = plt.Circle((0, 0), 150, fill=False)
        axs[2].add_artist(draw_circle)
        axs[2].grid(False)

        cbar = fig.colorbar(Xim, ax=axs[0], fraction=0.046, pad=0.04)
        cbar = fig.colorbar(Yim, ax=axs[1], fraction=0.046, pad=0.04)
        cbar = fig.colorbar(Zim, ax=axs[2], fraction=0.046, pad=0.04)
        plt.tight_layout()
        plt.show()

    def _plot_indices(self):  # pragma: no cover
        """
        handy debug plot routine to plot linear/distorted indices
        :return:
        """

        fig, axs = plt.subplots(nrows=2, ncols=2, figsize=[8, 8])
        axs[0, 0].plot(self.xj)
        axs[0, 0].set_title('xj')
        axs[0, 1].plot(self.yj)
        axs[0, 1].set_title('yj')
        axs[1, 0].plot(self.sk)
        axs[1, 0].set_title('sk')
        axs[1, 1].plot(self.tk)
        axs[1, 1].set_title('tk')

    def _plot_coords(self):  # pragma: no cover

        fig, axs = plt.subplots(1, 3)
        axs[0].plot(self._X_slice.flatten())
        axs[1].plot(self._Y_slice.flatten())
        axs[2].plot(self._Z_slice.flatten())
        plt.show()

    def _plot_images(self):  # pragma: no cover
        """
        debugging method to show input, output, and diff images
        :return:
        """
        vmin = 20
        vmax = 100
        fig, axs = plt.subplots(nrows=1, ncols=3)
        axs[0].imshow(self._image_to_correct, vmin=vmin, vmax=vmax)
        axs[1].imshow(self.outputImage, vmin=vmin, vmax=vmax)
        axs[2].imshow((abs(np.subtract(self._image_to_correct, self.outputImage))), vmin=vmin, vmax=vmax)
        axs[2].set_title('abs_difference')
        plt.tight_layout()

    # public methods

    def correct_all_images(self):
        """
        This will loop through and correct all IMA files in the input directory
        :return:
        """
        start_time = perf_counter()

        update_every_n_slices = 10
        if self.correct_through_plane:
            n_images_to_correct = self._n_dicom_files + (self.ImageArray.shape[1] - 2 * self._n_zero_pad)
        else:
            n_images_to_correct = self._n_dicom_files
        loop_axis = 2  # ImageArray always has slice last
        # nb: the below code allows us to wrap all the data into one 'itterable'; it also changes the view
        # such that we are looping over the slice direction
        zipped_data = zip(np.rollaxis(self.ImageArray, loop_axis),
                          np.rollaxis(self._X, loop_axis),
                          np.rollaxis(self._Y, loop_axis),
                          np.rollaxis(self._Z, loop_axis))

        i = 0
        self._image_array_corrected = np.zeros(self.ImageArray.shape)
        for array_slice, X, Y, Z in zipped_data:
            if i < self._n_zero_pad or i > self._n_dicom_files + self._n_zero_pad:
                # skip these files, they have no meaning anyway and we can't (easily) write them to dicom
                i += 1
                continue

            if i % update_every_n_slices == 0:
                print(printProgressBar(i-self._n_zero_pad, n_images_to_correct))
            self._image_to_correct = array_slice
            self._X_slice = X
            self._Y_slice = Y
            self._Z_slice = Z
            self._correct_image()
            self._image_array_corrected[:, :, i] = self.outputImage
            i += 1

        if self.correct_through_plane:

            if (np.round(self._ImageOrientationPatient) == [1, 0, 0, 0, 1, 0]).all():
                self._ImageOrientationPatient = [1, 1, 1, 1, 1, 1]
            elif np.round(self._ImageOrientationPatient == [0, 1, 0, 0, 0, -1]).all():
                self._ImageOrientationPatient = [2, 2, 2, 2, 2, 2]
            elif np.round(self._ImageOrientationPatient == [1, 0, 0, 0, 0, -1]).all():
                self._ImageOrientationPatient = [3, 3, 3, 3, 3, 3]
            else:
                raise NotImplementedError

            loop_axis = 1
            zipped_data = zip(np.rollaxis(self._image_array_corrected, loop_axis),
                              np.rollaxis(self._X, loop_axis),
                              np.rollaxis(self._Y, loop_axis),
                              np.rollaxis(self._Z, loop_axis))
            self._image_array_corrected2 = self._image_array_corrected.copy()
            j = 0
            self._Rows = np.rollaxis(self._image_array_corrected, loop_axis).shape[1]
            self._Cols = np.rollaxis(self._image_array_corrected, loop_axis).shape[2]
            n_slices = np.rollaxis(self._image_array_corrected, loop_axis).shape[0]
            for array_slice, X, Y, Z in zipped_data:
                if j < self._n_zero_pad or j > n_slices + self._n_zero_pad:
                    # skip these files, they have no meaning anyway and we can't (easily) write them to dicom
                    j += 1
                    continue

                if j % update_every_n_slices == 0:
                    print(printProgressBar(i+j-(self._n_zero_pad*4), n_images_to_correct))
                self._image_to_correct = array_slice
                self._X_slice = X
                self._Y_slice = Y
                self._Z_slice = Z
                self._correct_image()
                self._image_array_corrected[:, j, :] = self.outputImage

                j += 1

        execution_time = perf_counter() - start_time
        print(f'\ntotal time: {execution_time: 1.1f}s')
        print(f'mean time per slice = {execution_time / n_images_to_correct: 1.1}s')
        self._unpad_image_arrays()
        print(f'total correction time: {perf_counter() - start_time: 1.1f} s')

    def save_all_images(self, save_loc=None):
        """
        save corrected data as png

        :param save_loc: path to save data at.
        :type save_loc: string or path
        """
        plt.ioff()
        if save_loc is None:
            save_loc = self.ImageDirectory / 'corrected'
        save_loc = Path(save_loc)
        if not (self.ImageDirectory / save_loc).is_dir():
            save_loc.mkdir()

        loop_axis = 2
        zipped_data = zip(np.rollaxis(self.ImageArray, loop_axis),
                          np.rollaxis(self._image_array_corrected, loop_axis))
        i = 0
        for original_image, corrected_image in zipped_data:
            fig, axs = plt.subplots(nrows=1, ncols=2)

            # axs[0].imshow(self.pixel_array, extent=self._extent)
            im1 = axs[0].imshow(original_image)
            axs[0].set_title('Original Image')
            axs[0].set_aspect('equal')
            # axs[0].set_xlabel(self._row_label)
            # axs[0].set_ylabel(self._col_label)
            axs[0].grid(True)
            # fig.colorbar(im1, ax=axs[0])

            # axs[1].imshow(self.outputImage, extent=self._extent)
            im2 = axs[1].imshow(corrected_image)
            axs[1].set_title('Corrected Image')
            axs[1].set_aspect('equal')
            # axs[1].set_xlabel(self._row_label)
            # axs[1].set_ylabel(self._col_label)
            axs[1].grid(True)
            # fig.colorbar(im2, ax=axs[1])

            plt.tight_layout()
            plt.savefig(save_loc / (str(i) + '.png'), format='png')
            plt.close(fig)
            i += 1

    def save_all_images_as_dicom(self, save_loc=None):
        """
        save corrected data as dicom

        :param save_loc: path to save data at.
        :type save_loc: string or path
        """
        if self._image_array_corrected.min() < 0:
            self._image_array_corrected = self._zero_volume(self._image_array_corrected)
        if save_loc is None:
            save_loc = self.ImageDirectory / 'corrected_dcm'
        save_loc = Path(save_loc)
        if not save_loc.is_dir():
            save_loc.mkdir()
        loop_axis = 2
        zipped_data = zip(np.rollaxis(self.ImageArray, loop_axis),
                          np.rollaxis(self._image_array_corrected, loop_axis))
        i = 0
        for original_image, corrected_image in zipped_data:
            current_slice = pydicom.read_file(self.ImageDirectory / self._all_dicom_files[i])
            temp_dcm = current_slice.copy()
            temp_dcm.PixelData = np.uint16(corrected_image).tobytes()
            temp_dcm.save_as(save_loc / (str(i) + '.dcm'))
            i += 1


class KspaceDistortionCorrector(DistortionCorrectorBase):
    """
    :param NufftLibrary:
    """

    def __init__(self,NufftLibrary='finufft', pad=10, **kwds):

        super().__init__(pad=pad, **kwds)

        # Select nufft algorithm to use:
        if not NufftLibrary in ['pynufft', 'finufft', 'torchnufft']:
            raise NotImplementedError(f'Unknown NUFFTlibrary entered: {NufftLibrary}.'
                                      f'\nAllowed options are: pynufft, finufft, torchnufft')
        self.NUFFTlibrary = NufftLibrary  # pynufft, finufft, or torchnufft
        self._check_input_data()

    def _generate_Kspace_data(self):
        """
        Get the k space of the current slice data.
        :return:
        """
        self.k_space = fftshift(fft2(fftshift(self._image_to_correct)))

    def _correct_image(self):
        """
        List of steps required for each independant input slice.
        Under some circumstances some steps could be recycled for subsequent slices
        """

        self._image_shape = np.array(np.squeeze(self._X_slice).shape)
        coords_cartesian = np.array([self._X_slice.flatten(), self._Y_slice.flatten(), self._Z_slice.flatten()])
        coords_cartesian = pd.DataFrame(coords_cartesian.T, columns=['x', 'y', 'z'])
        self.coords = convert_cartesian_to_spherical(coords_cartesian)
        self._calculate_encoding_fields()
        self._generate_Kspace_data()
        self._calculate_encoding_indices()
        self._perform_least_squares_optimisation()

    def _fiNufft_Ax(self, x):
        """
        flatron instiute nufft
        Returns A*x
        equivalent to the 'notranpose' option in shanshans code
        self.xj and yj are non uniform nonuniform source points. they are essentially the encoding signals.
        self.sk and tk are uniform target points
        # """
        if x.dtype is not np.dtype('complex128'):
            x = x.astype('complex128')
        # y = nufft2d3(self.xj, self.yj, x, self.sk, self.tk, eps=1e-06, isign=-1)
        y = self.Nufft_Ax_Plan.execute(x, None)
        return y.flatten()

    def _fiNufft_Atb(self, x):
        """
        flatron instiute nufft
        This is to define the Nufft as a scipy.sparse.linalg.LinearOperator which can be used by the lsqr algorithm
        see here for explanation:
        https://stackoverflow.com/questions/48621407/python-equivalent-of-matlabs-lsqr-with-first-argument-a-function
        Returns A'*x
        equivalent to the 'tranpose' option in shanshans code
        """
        y = self.Nufft_Atb_Plan.execute(x, None)
        return y.flatten()

    def _perform_least_squares_optimisation(self):
        """
        From
        `Integrated Image Reconstruction and Gradient Nonlinearity Correction <https://www.ncbi.nlm.nih.gov/pmc/articles/PMC4390402/pdf/nihms629785.pdf>`_
        "A common strategy for reconstructing both fully- and undersampled MRI data is  penalized regression,
        which seeks to **produce the image** most likely to have produced the set of noisy measurements while potentially
        also satisfying some other expected properties (e.g., sparsity) (17). Since noise in MRI is Gaussian
        distributed, penalized least squares regression of the following general form is often employed.
        So:
        - the image is what we want to get back.
        - The noisy measurements are self.k_space
        - We use the NUFFT to compute the image most likely to have produced k_space, given the encoding fields that we computed.
        """
        fk1 = np.reshape(self.k_space, [self._Rows * self._Cols])
        StartingImage = None  # x0 for lsqr. can be overwritten for each option below.

        if self.NUFFTlibrary == 'finufft':
            self.Nufft_Ax_Plan = Plan(3, 2, 1, 1e-06, -1)
            self.Nufft_Ax_Plan.setpts(self.xj, self.yj, None, self.sk, self.tk)
            self.Nufft_Atb_Plan = Plan(3, 2, 1, 1e-06, 1)
            self.Nufft_Atb_Plan.setpts(self.sk, self.tk, None, self.xj, self.yj)
            A = LinearOperator((fk1.shape[0], fk1.shape[0]), matvec=self._fiNufft_Ax, rmatvec=self._fiNufft_Atb)
            StartingImage = self._image_to_correct.flatten().astype(complex)

        maxit = 20
        x1 = lsqr(A, fk1, iter_lim=maxit, x0=StartingImage)
        self.outputImage = abs(np.reshape(x1[0], [self._Rows, self._Cols]))


class ImageDomainDistortionCorrector(DistortionCorrectorBase):
    """

    """

    def _correct_image(self):
        """
        List of steps required for each independant input slice.
        Under some circumstances some steps could be recycled for subsequent slices
        """

        self._image_shape = np.array(np.squeeze(self._X_slice).shape)

        coords_cartesian = np.array([self._X_slice.flatten(), self._Y_slice.flatten(), self._Z_slice.flatten()])
        coords_cartesian = pd.DataFrame(coords_cartesian.T, columns=['x', 'y', 'z'])
        self.coords = convert_cartesian_to_spherical(coords_cartesian)
        self._calculate_encoding_fields()
        self._calculate_encoding_indices()

        x_n = np.arange(0, self._image_shape[0], 1)
        y_n = np.arange(0, self._image_shape[1], 1)
        pixel_interp = RectBivariateSpline(x_n, y_n, self._image_to_correct)

        yy, xx = np.meshgrid(y_n, x_n)
        xvector = (self.xn_dis_pixel + xx).flatten()
        yvector = (self.yn_dis_pixel + yy).flatten()

        _output_image = pixel_interp.ev(xvector, yvector)

        _output_image = np.where((xvector < 0) | (yvector < 0) | (xvector > self._image_shape[0]) |
                                 (yvector > self._image_shape[1]), 0, _output_image)

        self.outputImage = _output_image.reshape(self._image_shape)




from .Harmonics import SphericalHarmonicFit
import numpy as np

def calculate_harmonics(MagneticFields, n_order=8, norm=None):
    """
    This function is essentially a wrapper of convenience. Given a set of B fields and dicom_data,
    it will calculate and return the Gx, Gy, and Gz harmonics.
    Note that the gradient harmonics will be normalised to a gradient strengt of 1 mT/m

    :param MagneticFields: a pandas dataframe with columns ['x', 'y', 'z', 'B_Gx', 'B_Gy', 'B_Gz', 'B0']. B0 is optional.
        x, y, z are in mm and the fields are in any unit - the harmonics will reflect the units
    :type MagneticFields: pandas dataframe
    :param dicom_data: a dict containing required dicom data, such as that generated within the MarkerVolume class
    :type dicom_data: dict
    :param n_order: order of harmonic fit
    :type n_order: int, optional
    :param norm: Can pass a list of four valuues, the four returned harmonic lists will be normalised accordingly.
        e.g. norm = [2,2,2,2] will divide all harmonics by 2. This is useful so you can normalise the fields to some
        value
    :return: G_x_Harmonics, G_y_Harmonics, G_z_Harmonics, B0_Harmonics
    """

    if norm is None:
        norm = [1, 1, 1, 1]
    if not np.shape(norm)[0] == 4:
        raise ValueError('norm must be a list such as [2,2,2,2], where the elements correspond to Gx, Gy, Gz, and B0')

    # calculate harmonics

    # Gx
    GradXdata = MagneticFields[['x', 'y', 'z', 'B_Gx']]
    GradXdata = GradXdata.rename(
        columns={"B_Gx": "Bz"})  # spherical harmonics code expects to receieve one field called Bz
    G_x_Harmonics = SphericalHarmonicFit(GradXdata, n_order=n_order, r_outer=150, scale=norm[0])

    # Gy
    GradYdata = MagneticFields[['x', 'y', 'z', 'B_Gy']]
    GradYdata = GradYdata.rename(
        columns={"B_Gy": "Bz"})  # spherical harmonics code expects to receieve one field called Bz
    G_y_Harmonics = SphericalHarmonicFit(GradYdata, n_order=n_order, r_outer=150, scale=norm[1])

    # G_z
    GradZdata = MagneticFields[['x', 'y', 'z', 'B_Gz']]
    GradZdata = GradZdata.rename(
        columns={"B_Gz": "Bz"})  # spherical harmonics code expects to receieve one field called Bz
    G_z_Harmonics = SphericalHarmonicFit(GradZdata, n_order=n_order, r_outer=150, scale=norm[2])

    # B0
    try:
        B0_data = MagneticFields[['x', 'y', 'z', 'B0']]
        B0_data = B0_data.rename(
            columns={"B0": "Bz"})  # spherical harmonics code expects to receieve one field called Bz
        B0_Harmonics = SphericalHarmonicFit(B0_data, n_order=n_order, r_outer=150, scale=norm[3])
    except KeyError:
        B0_Harmonics = None

    return G_x_Harmonics, G_y_Harmonics, G_z_Harmonics, B0_Harmonics